# -*- coding: utf-8 -*-

from typing import List

from simmate.toolkit import Composition, Structure
from simmate.toolkit.creators.structure.prototypes import FromAflowPrototypes


def get_structures_from_prototypes(
    composition: Composition,
) -> List[Structure]:
    """
    Generates a list of structures by using prototypes as template. This function
    is a simple wrapper around the functionality in
    `simmate.toolkit.creators.structures.prototypes`. For more control over the
    structures generated, refer to this underlying creator class.
    """

    # generate input structures using all methods supported by the base class
    from_ordered = FromAflowPrototypes(
        composition,
        mode="ordered",
        allow_multiples=True,
    )
    from_exact = FromAflowPrototypes(
        composition,
        mode="exact",
    )

    # go through both list results and remove duplicate structures
    unique_structures = []
    for structure in from_ordered.structures + from_exact.structures:
        if structure not in unique_structures:
            unique_structures.append(structure)

    return unique_structures



# -*- coding: utf-8 -*-

"""
This module hosts common utilities for Simmate's workflow library. This includes
functions for grabbing all available workflows as well as dynamically loading
a workflow using its name.
"""

import os
import yaml
import shutil
import pkgutil
from typing import List
from importlib import import_module

from simmate import workflows
from simmate.utilities import get_directory, make_archive
from simmate.workflow_engine import Workflow


def get_workflow_types():
    """
    Grabs all workflow types, which is also all "Project Names" and all of the
    submodules listed in the `simmate.workflows` module.

    Gives names in the format with all lowercase and hypens, such as "relaxation"
    or "static-energy"
    """

    workflow_types = [
        submodule.name.replace("_", "-")
        for submodule in pkgutil.iter_modules(workflows.__path__)
        if submodule.name != "utilities"
    ]

    workflow_types.sort()

    return workflow_types


def get_list_of_workflows_by_type(
    flow_type: str,
    full_name: bool = True,
) -> List[str]:
    """
    Returns a list of all the workflows located in the given module.
    """

    # Make sure the type is supported
    workflow_types = get_workflow_types()
    if flow_type not in workflow_types:
        raise TypeError(
            f"{flow_type} is not allowed. Please use a workflow type from this"
            f" list: {workflow_types}"
        )

    # switch the naming convention from "flow-name" to "flow_name".
    flow_type_u = flow_type.replace("-", "_")

    # grab the relevent module
    flow_module = import_module(f"simmate.workflows.{flow_type_u}")

    # This loop goes through all attributes (as strings) of the workflow module
    # and select only those that are workflow or s3tasks.
    workflow_names = []
    for attr_name in dir(flow_module):
        attr = getattr(flow_module, attr_name)
        # OPTIMIZE: line below --> issubclass(attr, Workflow) --> raises error
        if hasattr(attr, "run_config") and attr.__name__ != "Workflow":
            # attr is now a workflow object (such as Relaxation__Vasp__Matproj)
            # and we can grab whichever name we'd like from it.
            if full_name:
                workflow_name = attr.name_full
            else:
                workflow_name = attr.name_preset
            # and add the name to our list
            workflow_names.append(workflow_name)
    # OPTIMIZE: is there a more efficient way to do this?

    workflow_names.sort()
    return workflow_names


def get_list_of_all_workflows() -> List[str]:
    """
    Returns a list of all the workflows of all types.

    This utility was make specifically for the CLI where we print out all
    workflow names for the user.
    """

    workflow_names = []
    for flow_type in get_workflow_types():
        workflow_names += get_list_of_workflows_by_type(flow_type)

    return workflow_names


def get_workflow(
    workflow_name: str,
    precheck_flow_exists: bool = False,
    print_equivalent_import: bool = False,
) -> Workflow:
    """
    This is a utility for that grabs a workflow from the simmate workflows.

    This is typically used when we need to dynamically import a workflow, such
    as when calling a workflow through the command-line. We recommend avoiding
    this function when possible, and instead directly import your workflow.

    For example...
    ``` python
    from simmate.workflows.utilities import get_workflow

    matproj_workflow = get_workflow("static-energy.vasp.matproj")
    ```

    ...does the same exact thing as...
    ``` python
    from simmate.workflows.static_energy import StaticEnergy__Vasp__Matproj
    ```

    Note the naming of workflows therefore follows the format:
    ``` python

    # an example import of some workflow
    from simmate.workflows.example_module import example_workflow_class

    ```

    #### Parameters

    - `workflow_name`:
        Name of the workflow to load. Examples include "relaxation.vasp.matproj",
        "static-energy.vasp.quality01", and "diffusion.vasp.all-paths"

    - `precheck_flow_exists`:
        Whether to check if the workflow actually exists before attempting the
        import. Note, this requires loading all other workflows in order to make
        this check, so it slows down the function substansially. Defaults to false.

    - `print_equivalent_import`:
        Whether to print a message indicating the equivalent import for this
        workflow. Typically this is only useful for beginners using the CLI.
        Defaults to false.
    """

    # make sure we have a proper workflow name provided.
    # This is optional because it is slow and loads all other workflows, rather
    # than just the one we are interested in.
    if precheck_flow_exists:
        allowed_workflows = get_list_of_all_workflows()
        if workflow_name not in allowed_workflows:
            raise ModuleNotFoundError(
                "The workflow you provided isn't known. Make sure you don't have any "
                "typos! If you want a list of all available workflows, use the command "
                "`simmate workflows list-all`. You can also interactively explore "
                "workflows with `simmate workflows explore`"
            )

    # parse the workflow name. (e.g. static-energy.vasp.mit --> static_energy + vasp + mit)
    project_name, calculator_name, preset_name = workflow_name.replace("-", "_").split(
        "."
    )

    # Combine the names into the full class name
    # (e.g. static_energy + vasp + mit --> StaticEnergy__Vasp__Mit)
    workflow_class_name = "__".join(
        [
            n.title().replace("_", "")
            for n in [project_name, calculator_name, preset_name]
        ]
    )

    # The naming convention matches the import path
    # BUG: What about user workflows...? Should I try each custom app import?
    workflow_module = import_module(f"simmate.workflows.{project_name}")

    # If requested, print a message indicating the import we are using
    if print_equivalent_import:
        print(
            "Using: \n\n\t"
            f"from simmate.workflows.{project_name} import {workflow_class_name} \n\n"
            "You can find the source code for this workflow in the follwing module: \n\n\t"
            f"simmate.calculators.{calculator_name}.workflows.{project_name}"
        )

    # and import the workflow
    workflow = getattr(workflow_module, workflow_class_name)

    return workflow


def load_results_from_directories(
    base_directory: str = ".",
):
    """
    Goes through a given directory and finds all "simmate-task-" folders and zip
    archives present. The simmate_metadata.yaml file is used in each of these
    to load results into the database. All folders will be converted to archives
    once they've been loaded.

    #### Parameters

    - `base_directory`:
        The main directory that will contain folders to archive. Defaults to the
        working directory.
    """
    # load the full path to the desired directory
    directory = get_directory(base_directory)

    # grab all files/folders in this directory and then limit this list to those
    # that are...
    #   1. folders
    #   2. start with "simmate-task-"
    #   3. haven't been modified for at least time_cutoff
    foldernames = [
        os.path.join(directory, f)
        for f in os.listdir(directory)
        if "simmate-task-" in os.path.basename(f)
    ]

    # Now go through this list and archive the folders that met the criteria
    # and load the data from each.
    for foldername in foldernames:

        # Print message for monitoring progress.
        # TODO: consider switching to tqdm
        print(f"Loading data from... \n\t{foldername}")

        # There may be many folders that contain failed or incomplete data.
        # We don't want those to prevent others from being loaded so we put
        # everything in a try/except.
        try:

            # If we have a zip file, we need to unpack it before we can read results
            if not os.path.isdir(foldername):
                shutil.unpack_archive(
                    filename=foldername,
                    extract_dir=directory,
                )
                # remove the ".zip" ending for our folder
                foldername = foldername.removesuffix(".zip")

            # Grab the metadata file which tells us key information
            filename = os.path.join(foldername, "simmate_metadata.yaml")
            with open(filename) as file:
                metadata = yaml.full_load(file)

            # see which workflow was used -- which also tells us the database table
            workflow_name = metadata["workflow_name"]
            workflow = get_workflow(workflow_name)

            # now load the data
            results_db = workflow.database_table.from_directory(foldername)

            # use the metadata to update the other fields
            results_db.source = metadata["source"]
            results_db.prefect_flow_run_id = metadata["prefect_flow_run_id"]

            # note the directory might have been moved from when this was originally
            # ran vs where it is now. Therefore, we update the folder location here.
            results_db.directory = foldername

            # Now save the results and convert the folder to an archive
            results_db.save()
            make_archive(foldername)

            print("\tSuccessful.")

        except:
            print("\tFailed.")


def get_unique_parameters() -> List[str]:
    """
    Returns a list of all unique parameter names accross all workflows.

    This utility is really just to help developers ensure they are covering
    all cases when implementing new features, so this function isn't actually
    called elsewhere.
    """

    flownames = get_list_of_all_workflows()
    unique_parameters = []
    for flowname in flownames:
        workflow = get_workflow(flowname)
        for parameter in workflow.parameter_names:
            if parameter not in unique_parameters and parameter not in [
                "kwargs",
                "cls",
            ]:
                unique_parameters.append(parameter)

    # for consistency, we sort these alphabetically
    unique_parameters.sort()

    return unique_parameters



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import table_column, Structure, Thermodynamics


class MatprojStructure(Structure, Thermodynamics):
    """
    Crystal structures from the [Materials Project](https://materialsproject.org/)
    database.

    Currently, this table only stores strucure and thermodynamic information,
    but the Materials Project has much more data available via their
    [REST API](https://github.com/materialsproject/api) and website.
    """

    class Meta:
        app_label = "third_parties"

    base_info = ["id", "structure_string", "energy"]
    source = "Materials Project"
    source_doi = "https://doi.org/10.1063/1.4812323"
    remote_archive_link = "https://archives.simmate.org/MatProjStructure-2022-01-26.zip"

    id = table_column.CharField(max_length=25, primary_key=True)
    """
    The id used to represent the structure (ex: "mp-12345")
    """

    @property
    def external_link(self) -> str:
        """
        URL to this structure in the Materials Project website.
        """
        # All Materials Project structures have their data mapped to a URL in
        # the same way. For example...
        #   https://materialsproject.org/materials/mp-12345/
        return f"https://materialsproject.org/materials/{self.id}/"



# -*- coding: utf-8 -*-

from pymatgen.transformations.standard_transformations import (
    PerturbStructureTransformation,
)

from simmate.toolkit.transformations.base import Transformation


class CoordinatePerturbation(Transformation):

    # random perturbation of sites was removed from USPEX because primarily made more defective + lower energy structures
    # https://uspex-team.org/static/file/USPEX-LargeComplexSystems-2010.pdf

    # note, the function Structure.pertrub displaces each site at an equal distance
    # this function on the other hand displaces at a variable distance selected from a guassian distribution between min/mix
    # this transformation is almost guaranteed to give a structure with P1 symmetry

    io_scale = "one_to_one"
    ninput = 1
    use_multiprocessing = False

    def __init__(self, min_displace=0, max_displace=1):  #!!! whats a good displacement?

        self.perturb_object = PerturbStructureTransformation(min_displace, max_displace)

    def apply_transformation(self, structure):

        new_structure = self.perturb_object.apply_transformation(structure)

        return new_structure



# -*- coding: utf-8 -*-

from simmate.database.base_data_types import DiffusionAnalysis

(
    MITDiffusionAnalysis,
    MITMigrationHop,
    MITMigrationImage,
) = DiffusionAnalysis.create_subclasses("MIT", module=__name__)



# -*- coding: utf-8 -*-

import os

import numpy

from sklearn.model_selection import train_test_split
from django_pandas.io import read_frame

from simmate.toolkit import Structure, Composition
from simmate.utilities import get_directory


class DeepmdDataset:
    """
    This class works simply as a converter. You provide it a list of
    IonicStepStructures that have forces and energies, and it will create a
    deepmd input for you.

    The input consists of 2 "type" files and then a subfolder made of 4 files for
    the actaul data. An example folder looks like this:
        type.raw
        type_map.raw
        set.000
            box.npy
            coord.npy
            energy.npy
            force.npy

    The "type" files are very simple. "type_map.raw" is just a list of elements
    and the "type.raw" is our composition listing what each site is. For example,
    H20 (water) would just be...
        type_map.raw --> H O
        type.raw --> 0 0 1

    These .npy are "numpy" files, where the data included in each is...
      box = lattice matrixes
      coord = site's cartesian coordinates
      energy = calculated energies
      force = calculated forces for sites

    All data is collapsed to 1D arrays. For example, this means the 3x3 lattice matrix
    becomes a 1x9 list of numbers and the entire file is a list of matricies in this
    format. The same is done to forces, coords, and energies.

    Note this is the folder/file format for EACH composition. So this class creates
    many of these folders if you provide input structures that have varying
    compositions (such as CoO2, Co2O4, Co3O6, etc.).

    Further, we split the input structures into training and test datasets. So
    the final folder setup will look like...
    deepmd_data
        CoO2_train
        CoO2_test
        Co2O4_train
        Co2O4_test
        << etc. >>

    All data required is available from an IonicStepStructure in our database, so
    this is our current input format.
    """

    # TODO: currently we use the IonicStepStructure from our relaxation database
    # but in the future this should be a core class that extends the pymatgen
    # Structure object. So for the moment, the ionic_step_structures can be
    # provide like so:
    #
    # from simmate.database import connect
    # from simmate.database.workflow_results import MITRelaxationStructure
    # ionic_step_structures = MITRelaxationStructure.objects.filter(
    #     energy__isnull=False, site_forces__isnull=False
    # ).all()
    #
    # from simmate.calculators.deepmd.inputs.set import DeepmdSet
    # DeepmdSet.to_folder(ionic_step_structures)
    #

    @staticmethod
    def to_file(
        ionic_step_structures,
        directory="deepmd_data",
        test_size=0.2,
    ):

        # Grab the path to the desired directory and create it if it doesn't exist
        directory = get_directory(directory)

        # convert the ionic_step_structures queryset to a pandas dataframe
        structures_dataframe = read_frame(ionic_step_structures)

        # because we are using the database model, we first want to convert to
        # pymatgen structures objects and add a column to the dataframe for these
        #
        #   structures_dataframe["structure"] = [
        #       structure.to_toolkit() for structure in ionic_step_structures
        #   ]
        #
        # BUG: the read_frame query creates a new query, so it may be a different
        # length from ionic_step_structures. For this reason, we can't iterate
        # through the queryset like in the commented out code above. Instead,
        # we need to iterate through the dataframe rows.
        # See https://github.com/chrisdev/django-pandas/issues/138 for issue
        structures_dataframe["structure"] = [
            Structure.from_str(s.structure_string, fmt="POSCAR")
            for _, s in structures_dataframe.iterrows()
        ]

        # split the structures into test and training sets randomly
        dataframe_train, dataframe_test = train_test_split(
            structures_dataframe,
            test_size=test_size,
        )

        # The process for creating files is the same for both the test and training
        # sets, where the only difference is the folder ending we add to each. Other
        # methods (such as creating the input.json for DeePMD) require the names
        # of the folders created here -- so we also store them in lists to return
        # at the end of the function too.
        folders_train = []
        folders_test = []
        for folder_suffix, dataframe_set, folder_list in [
            ("train", dataframe_train, folders_train),
            ("test", dataframe_test, folders_test),
        ]:

            # grab a list of the unique compositions in our set of structures
            unique_compositions = dataframe_set.formula_full.unique()

            # for each composition, we want to filter off those structures and
            # then write them to a individual folder for deepmd.
            # note these compositions are just strings right now.
            for composition_str in unique_compositions:

                # convert to a pymatgen composition object
                composition = Composition(composition_str)

                # Let's establish where this folder will be and also store it
                composition_directory = os.path.join(
                    directory, composition_str + "_" + folder_suffix
                )
                folder_list.append(composition_directory)
                # this creates the directory or grabs the full path
                composition_directory = get_directory(composition_directory)

                # first let's write the type_map file, which is just a list of elements
                mapping_filename = os.path.join(composition_directory, "type_map.raw")
                with open(mapping_filename, "w") as file:
                    for element in composition:
                        file.write(str(element) + "\n")

                # Now we can write the type file while also establish the mapping.
                # Note the mapping is just the index (0, 1, 2, ...) of each element.
                type_filename = os.path.join(
                    directory, composition_str + "_" + folder_suffix, "type.raw"
                )
                with open(type_filename, "w") as file:
                    for mapping_value, element in enumerate(composition):
                        for i in range(int(composition[element])):
                            file.write(str(mapping_value) + "\n")

                # Now we need the relevent structures. Let's filter off the
                # structures that have this composition for us to use below
                dataframe_subset = dataframe_set[
                    dataframe_set["formula_full"] == composition_str
                ]

                # We iterate through each structure (and its data) to compile everything
                # into the deepmd list format.
                lattices = []
                coordinates = []
                forces = []
                energies = []
                # Note the "_," here is because row index is returned but we dont need it
                for _, structure_data in dataframe_subset.iterrows():

                    # grab the pymatgen structure
                    structure = structure_data.structure

                    # flatten the lattice matrix to a 1D array and add to our list
                    lattice_flat = structure.lattice.matrix.flatten()
                    lattices.append(lattice_flat)

                    # flatten the cartesian coordinates to a 1D array and add to our list
                    coords_flat = structure.cart_coords.flatten()
                    coordinates.append(coords_flat)

                    # flatten the forces to a 1D array and add to our list
                    forces_flat = numpy.array(structure_data.site_forces).flatten()
                    forces.append(forces_flat)

                    # the energy is just single value so we can add it to our list
                    energies.append(structure_data.energy)

                # Now we want to convert all lists to numpy.
                lattices = numpy.array(lattices)
                coordinates = numpy.array(coordinates)
                forces = numpy.array(forces)
                energies = numpy.array(energies)
                # Note, the dtype=object can be added as a kwarg here in order to clear
                # a warning that prints when we have arrays of different lengths. This
                # happens when we have structures with different nsites. However, this
                # causes errors with DeepMD which can load object type arrays.

                # for now we assume the dataset is written to set.000, so we
                # make that folder first.
                set_directory = os.path.join(composition_directory, "set.000")
                set_directory = get_directory(set_directory)
                # now write our numpy files to the folder specified
                for filename, filedata in [
                    ("box.npy", lattices),
                    ("coord.npy", coordinates),
                    ("energy.npy", energies),
                    ("force.npy", forces),
                ]:
                    filename_full = os.path.join(set_directory, filename)
                    with open(filename_full, "wb") as file:
                        numpy.lib.format.write_array(fp=file, array=filedata)

        # all of the folders have been created and we return a list of where
        return folders_train, folders_test



# -*- coding: utf-8 -*-

from tqdm import tqdm

from simmate.toolkit import Structure as ToolkitStructure
from simmate.database.base_data_types import DatabaseTable, table_column
from simmate.utilities import get_chemical_subsystems

from pymatgen.analysis.phase_diagram import PDEntry
from pymatgen.analysis.phase_diagram import PhaseDiagram


class Thermodynamics(DatabaseTable):
    class Meta:
        abstract = True

    base_info = ["energy"]

    energy = table_column.FloatField(blank=True, null=True)
    """
    The calculated total energy. Units are in eV.
    """

    energy_per_atom = table_column.FloatField(blank=True, null=True)
    """
    The `energy` divided by `nsites`. Units are in eV.
    """

    # TODO: These columns below should be updated on a schedule. I'd run
    # a prefect flow every night that makes sure all values are up to date.

    energy_above_hull = table_column.FloatField(blank=True, null=True)
    """
    The hull energy (aka "stability") of this structure compared to all other
    structures in this database table. Units are in eV.
    """

    is_stable = table_column.BooleanField(blank=True, null=True)
    """
    Whether `energy_above_hull` is 0 -- if so, this is considered stable.
    """

    decomposes_to = table_column.JSONField(blank=True, null=True)
    """
    If `energy_above_hull` is above 0, these are the compositions that the 
    structure will decompose to (e.g. ["Y2C", "C", "YF3"])
    """

    formation_energy = table_column.FloatField(blank=True, null=True)
    """
    The formation energy of the structure relative to all other
    structures in this database table. Units are in eV.
    """

    formation_energy_per_atom = table_column.FloatField(blank=True, null=True)
    """
    The `formation_energy` divided by `nsites`.
    """

    # Other fields to consider
    # equilibrium_reaction_energy_per_atom
    # energy_uncertainy_per_atom
    # energy_per_atom_uncorrected
    # decompose_amount (to each decomp type)

    """ Properties """
    # These are some extra fields to consider adding
    # energy_type = ... (DFT, machine-learned, LJ, etc.)
    # calc_types = (list) https://github.com/materialsproject/emmet/blob/main/emmet-core/emmet/core/vasp/calc_types/utils.py
    # psudeopotential
    #     functional(PBE)
    #     label(Y_sv)
    #     pot_type(PAW)
    # type(GGA,GGAU,HF)
    # is_hubbard

    @classmethod
    def _from_toolkit(
        cls,
        structure: ToolkitStructure,
        energy: float = None,
        as_dict: bool = False,
    ):

        # TODO: should structure be optional?

        # Given energy, this function builds the rest of the required fields
        # for this class as an object (or as a dictionary).
        data = (
            dict(
                energy=energy,
                energy_per_atom=energy / structure.num_sites,
                energy_above_hull=None,
                is_stable=None,
                decomposes_to=None,
            )
            if energy
            else {}
        )

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return data if as_dict else cls(**data)

    @classmethod
    def update_chemical_system_stabilities(cls, chemical_system: str):

        # NOTE: I assume we are using a Child(Structure, Thermodynamics)

        # if we have a multi-element system, we need to include subsystems as
        # well. ex: Na --> Na, Cl, Na-Cl
        subsystems = get_chemical_subsystems(chemical_system)

        # grab all entries for this chemical system
        entries = (
            cls.objects.filter(
                chemical_system__in=subsystems,
                energy__isnull=False,  # only completed calculations
            )
            .only("energy", "formula_full")
            .all()
        )

        # convert to pymatgen PDEntries and build into PhaseDiagram object
        entries_pmg = [PDEntry(entry.formula_full, entry.energy) for entry in entries]
        phase_diagram = PhaseDiagram(entries_pmg)

        # now go through the entries and update stability values
        for entry, entry_pmg in zip(entries, entries_pmg):

            decomp, hull_energy = phase_diagram.get_decomp_and_e_above_hull(entry_pmg)

            entry.energy_above_hull = hull_energy

            entry.is_stable = True if hull_energy == 0 else False

            # OPTIMIZE: I would like this to point to another entry specifically
            # but this will take more work.
            entry.decomposes_to = (
                [d.composition.formula for d in decomp] if hull_energy != 0 else []
            )

            entry.formation_energy = phase_diagram.get_form_energy(entry_pmg)
            entry.formation_energy_per_atom = phase_diagram.get_form_energy_per_atom(
                entry_pmg
            )

        # Now that we updated our objects, we want to collectively update them
        cls.objects.bulk_update(
            objs=entries,
            fields=[
                "energy_above_hull",
                "is_stable",
                "decomposes_to",
                "formation_energy",
                "formation_energy_per_atom",
            ],
            # updating extremely large systems (>3k structures) can cause this
            # to time-out and crash. We therefore update in batches of 500
            batch_size=500,
        )

    @classmethod
    def update_all_stabilities(cls):

        # grab all unique chemical systems
        chemical_systems = cls.objects.values_list(
            "chemical_system", flat=True
        ).distinct()

        # Now  go through each and run the analysis.
        # OPTIMIZE: Some systems will be analyzed multiple times. For example,
        # C would be repeatedly updated through C, C-O, Y-C-F, etc.
        for chemical_system in tqdm(chemical_systems):
            try:
                cls.update_chemical_system_stabilities(chemical_system)
            except ValueError as exception:
                print(f"Failed for {chemical_system} with error: {exception}")

        # BUG: can't use parallel=True as an input
        # Because different systems may need to update a single one at the same
        # time, errors will be thrown due to row locking. For example, Y-C and
        # Sc-C system might both try to update a C structure at the same time
        # and one will throw an error.
        #
        # from simmate.configuration.dask import batch_submit
        #
        # batch_submit(
        #     function=cls.update_chemical_system_stabilities,
        #     args_list=chemical_systems,
        #     batch_size=1000,
        # )



# -*- coding: utf-8 -*-

import os

import numpy

from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.outputs import Oszicar
from simmate.workflow_engine import ErrorHandler


class PositiveEnergy(ErrorHandler):
    """
    Check if a run has positive absolute energy.
    If so, we trying changeing ALGO to Normal or alternatively halve the POTIM.
    """

    # run this while the VASP calculation is still going
    is_monitor = True

    def check(self, directory: str) -> bool:
        """
        Check for error in the specified directory. Note, we assume that we are
        checking the OSZICAR file. If that file is not present, we say that there
        is no error because another handler will address this.
        """

        # We check for this error in the OSZICAR because it's the smallest file
        # that will tell us energies -- and therefore the fastest to read.
        filename = os.path.join(directory, "OSZICAR")

        # check to see that the file is there first
        if os.path.exists(filename):

            # then load the file's data
            oszicar = Oszicar(filename)

            # before we check the final energy, we first need to make sure at
            # least one ionic step is present. If not, there isn't an error yet
            if not oszicar.ionic_steps:
                return False

            # the final ionic step may still not have an energy because it's
            # still going. We therefore check the most recent energy and not
            # the final energy.
            for ionic_step in reversed(oszicar.ionic_steps):
                # where ionic steps are not complete, we have numpy.NAN values
                if not numpy.isnan(ionic_step["energy_sigma_zero"]):
                    break

            # If this energy is positive, we have an error.
            if ionic_step["energy_sigma_zero"] > 0:
                return True

        # if the file doesn't exist OR if we get through our checks above, then
        # we can say there is no error.
        return False

    def correct(self, directory: str) -> str:
        """
        Perform corrections based on the INCAR.
        """
        # Note "error" here is just True because there is no variation in this ErrorHandler.
        # This value isn't used in fixing the Error anyways.

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # check what the current ALGO is. If it's not set, that means it's using
        # the default which is "Normal".
        current_algo = incar.get("ALGO", "Normal")

        # If the current algo is *not* Normal, then switch it to that
        if current_algo != "Normal":

            # Set the new value
            incar["ALGO"] = "Normal"

            # rewrite the new INCAR file
            incar.to_file(incar_filename)

            # return the description of what we did
            return f"switched ALGO from {current_algo} to Normal"

        # if the algo is already Normal, we try reducing the POTIM by half
        else:

            # check what the current POTIM is. If it's not set, that means it's using
            # the default which is 0.5.
            current_potim = incar.get("POTIM", 0.5)

            # half the potim to try fixing the error
            new_potim = current_potim / 2.0

            # Set the new value
            incar["POTIM"] = new_potim

            # rewrite the new INCAR file
            incar.to_file(incar_filename)

            # return the description of what we did
            return f"halved POTIM from {current_potim} to {new_potim}"



# -*- coding: utf-8 -*-

import os
import json

from pymatgen.io.vasp.outputs import Outcar

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Brmix(ErrorHandler):
    """
    This handler addresses changes in charge density during a SCF loop. There are
    a series of fixes that depend on the type and state of the calculation being
    ran, so be sure to read the stages below.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["BRMIX: very serious problems"]

    def check(self, directory: str) -> bool:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # if NELECT is in the INCAR, that means we are running
        # a charged calculation (e.g. defects). If this is the
        # case, then we want to ingore a change in electron
        # density (brmix) and move on to checking the next error.
        if "NELECT" in incar:
            return False

        # otherwise check the file as usual for the error
        return super().check(directory)

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # load the error-count file if it exists
        error_count_filename = os.path.join(directory, "simmate_error_counts.json")
        if os.path.exists(error_count_filename):
            with open(error_count_filename) as error_count_file:
                error_counts = json.load(error_count_file)
        # otherwise we are starting with an empty dictionary
        else:
            error_counts = {}

        # The fix is based on the number of times we've already tried to
        # fix brmix. So let's make sure it's in our error_count dictionary.
        # If it isn't there yet, set the count to 0 and we'll update it below.
        error_counts["brmix"] = error_counts.get("brmix", 0)

        # The KGAMMA mode is relevent in a number of fixes below, so we
        # just grab that here. The default value is True.
        current_kgamma = incar.get("KGAMMA", True)

        # check if there is a valid OUTCAR
        if error_counts["brmix"] == 0:
            outcar_filename = os.path.join(directory, "OUTCAR")
            try:
                assert Outcar(outcar_filename).is_stopped is False
            except Exception:
                # if the OUTCAR isn't valid, we want to skip the first attempted
                # correction below. We do this by adding 1 to our error count.
                error_counts["brmix"] += 1

        # we don't return the correction immediately because we need to rewrite some
        # files below first. Therefore we just store the correction as a variable.

        if error_counts["brmix"] == 0:
            # If the OUTCAR is valid - simply rerun the job and increment
            # error count for next time.
            incar["ISTART"] = 1
            correction = "switched ISTART to 1"
            error_counts["brmix"] += 1

        elif error_counts["brmix"] == 1:
            # Use Kerker mixing w/default values for other parameters
            incar["IMIX"] = 1
            correction = "switched IMIX to 1"
            error_counts["brmix"] += 1

        elif error_counts["brmix"] == 2 and current_kgamma == True:
            # switch to Monkhorst and turn off Kerker mixing (IMIX=1)
            incar["KGAMMA"] = False
            incar.pop("IMIX", None)
            correction = "removed any IMIX tag and switched KGAMMA to False"
            error_counts["brmix"] += 1

        elif error_counts["brmix"] in [2, 3] and current_kgamma == False:
            # switch to Gamma and turn off Kerker mixing (IMIX=1)
            incar["KGAMMA"] = True
            incar.pop("IMIX", None)
            correction = "removed any IMIX tag and switched KGAMMA to True"
            error_counts["brmix"] += 1

            # TODO: I'm not sure what Custodian did here, so I haven't adapted
            # it yet. How can the number of kpoints be less than 1...?
            #
            # if vi["KPOINTS"].num_kpts < 1:
            #     all_kpts_even = all(bool(n % 2 == 0) for n in vi["KPOINTS"].kpts[0])
            #     if all_kpts_even:
            #         new_kpts = (tuple(n + 1 for n in vi["KPOINTS"].kpts[0]),)
            #         actions.append(
            #             {
            #                 "dict": "KPOINTS",
            #                 "action": {"_set": {"kpoints": new_kpts}},
            #             }
            #         )

        else:

            # Try turning off symmetry and using a Gamma-packed grid
            incar["ISYM"] = 0
            incar["KGAMMA"] = True
            correction = "switched ISYM to 0 and KGAMMA to True"

            # Check the current ICHARG setting, where default is 0
            # If the ICHARG is less than 10, then we want to delete the CHGCAR
            # and WAVECAR to ensure the next run is a clean start.
            current_icharg = incar.get("ICHARG", 0)
            if current_icharg < 10:
                os.remove(os.path.join(directory, "CHGCAR"))
                os.remove(os.path.join(directory, "WAVECAR"))
                correction += " and deleted CHGCAR and WAVECAR"
            # BUG: why doesn't custodian add an attempt here?
            # error_counts["brmix"] += 1

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        # rewrite the new error count file
        with open(error_count_filename, "w") as file:
            json.dump(error_counts, file)

        # now return the correction made for logging
        return correction



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class TetrahedronMesh(ErrorHandler):
    """
    This handler addresses a series of error messages that are all caused by
    having incompatible tetrahedral mesh. In some cases, increasing the k-point
    mesh will fix the issue, and in other cases, we'll simply turn off the
    tetrahedron method.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = [
        "Tetrahedron method fails",
        "Fatal error detecting k-mesh",
        "Fatal error: unable to match k-point",
        "Routine TETIRR needs special values",
        "Tetrahedron method fails (number of k-points < 4)",
        "BZINTS",
        # This error is separate from the tetrahedral methods above, but it has
        # the same associated fix -- so we include it here.
        "DENTET",
    ]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # check what the current KSPACING is. If it's not set, that means we're using
        # the default which is 0.5.
        current_kspacing = incar.get("KSPACING", 0.5)

        # If KSPACING isn't the default value, we try decreasing KSPACING
        # by 20% in each direction. which approximately doubles the number
        # of kpoints.
        if current_kspacing != 0.5:
            new_kspacing = current_kspacing * 0.8
            incar["KSPACING"] = new_kspacing
            # rewrite the INCAR with new settings
            incar.to_file(incar_filename)
            return f"switched KSPACING from {current_kspacing} to {new_kspacing}"

        # otherwise we try changing the smearing method to guassian
        else:
            incar["ISMEAR"] = 0
            incar["SIGMA"] = 0.05
            # rewrite the INCAR with new settings
            incar.to_file(incar_filename)
            return "switched ISMEAR to 0 and SIGMA to 0.05"



# -*- coding: utf-8 -*-

import pandas as pd
from numpy.random import choice, shuffle
import itertools

from simmate.toolkit import Structure, Composition

from simmate.toolkit.symmetry.wyckoff import loadWyckoffData
from simmate.toolkit.creators.vector import UniformlyDistributedVectors
from simmate.toolkit.creators.sites.random_wyckoff import (
    asymmetric_unit_boundries,
)  #!!! this will move to simmate.toolkit.symmetry in the future
from simmate.toolkit.creators.lattice import RSLSmartVolume

# from pymatdisc.core.estimate import distance_matrix # !!! TODO

##############################################################################


class StructureCreator:
    def __init__(
        self,
        composition,
        spacegroup_include=range(1, 231),
        spacegroup_exclude=[],
        max_total_attempt=1e6,
        max_large_mult_attempt=400,
        max_dist_attempt=100,
    ):

        # save general input
        self.max_large_mult_attempt = max_large_mult_attempt
        self.max_dist_attempt = max_dist_attempt
        self.max_total_attempt = max_total_attempt
        self.composition = composition
        self.spacegroup_exclude = spacegroup_exclude
        self.spacegroup_include = spacegroup_include

        # list of spacegroups that can be used
        self.space_group_options = [
            sg for sg in spacegroup_include if sg not in spacegroup_exclude
        ]
        # assign lattice gen method
        self.lattice_generator = RSLSmartVolume(composition)
        # assign udv to spacegroups
        self.vector_creator = {}
        for sg in self.space_group_options:
            aub = asymmetric_unit_boundries(sg)
            udv = UniformlyDistributedVectors(extra_conditions=aub)
            self.vector_creator[sg] = udv
        # load wyckoff data
        self.wy_data = loadWyckoffData()

    def new_structure(self, spacegroup=None, lattice=None, species=None, coords=None):

        # shuffle compostion order
        elements = [e for e in self.composition]
        shuffle(elements)
        comp_dict = {}

        for i in elements:
            comp_dict[i] = self.composition[i]
        # create new ordered composition from dictionary
        new_composition = Composition.from_dict(comp_dict)
        # create cutoff matrix for composition
        cutoff_matrix = distance_matrix(new_composition, radius_method="atomic")

        if not spacegroup:
            spacegroup = choice(self.space_group_options)

        udv = self.vector_creator[spacegroup]

        if not lattice:
            struct_lattice = self.lattice_generator.new_lattice(spacegroup)

        spacegroup_data = self.wy_data.query("SpaceGroup==@spacegroup")
        sg_data = spacegroup_data.values

        # list to hold every wy_site added
        master_wy_list = []
        species_list = []
        coords_list = []
        # test_struct = Structure(struct_lattice, species_list, coords_list)

        for element in new_composition:
            # create list to hold wy sites for element
            element_sites = []
            # list to hold multiplicity values
            mults = []
            # while loop check
            master_check = False
            large_mult_attempt = 0
            total_attempt_count = 0
            while not master_check:  # while not check and (attempts <= ...)

                #!!! add an attempt
                total_attempt_count += 1
                if total_attempt_count == self.max_total_attempt:
                    self.space_group_options.remove(spacegroup)
                    return False

                # grab a random wy site
                wy_site = sg_data[choice(len(sg_data))]
                wy_letter = wy_site[3]  # 3 = index of letter
                wy_coord = wy_site[5]  # 5 = index of coordinate
                wy_mult = wy_site[1]  # 1 = site multiplicity

                # check site availability
                avail_check = True
                if wy_site[6] == 1.0:  # 6 = site availability
                    # see if this wy_letter has been used by the current element yet
                    if wy_letter in element_sites:
                        avail_check = False
                    # see if any other elements used the wy_site already
                    for species in master_wy_list:
                        if wy_letter in species:
                            avail_check = False
                            break  # no need to look at any other species
                    if avail_check == False:
                        continue  # restart while loop
                # else: all other sites are assumed to have infinite availability

                # check multiplicity
                mult_check = False
                mults_total = sum(mults) + wy_mult
                if mults_total == self.composition[element]:
                    element_sites.append(wy_letter)
                    mults.append(wy_site[1])  # 1 = multiplicity of wy_site
                    mult_check = True  #!!! is this the correct placement? Even before distance check? -Jack
                elif mults_total < self.composition[element]:
                    mults.append(wy_site[1])
                    element_sites.append(wy_letter)
                elif mults_total > self.composition[element]:
                    large_mult_attempt += 1
                    if (
                        large_mult_attempt == self.max_large_mult_attempt
                        and element_sites
                    ):
                        element_sites.pop()
                        mults.pop()
                    continue  # restart while loop

                # check if wy_letter can be successfully added to structure
                dist_check = False
                dist_attempt = 0

                while not dist_check:

                    # add an attempt
                    dist_attempt += 1

                    # see if we have maxed out our number of attempts
                    if dist_attempt == self.max_dist_attempt:
                        break  # this breaks the while loop with dist_check = False

                    # make the new vector coords
                    # note that these value will be used to eval the wy_site coords
                    x, y, z = udv.new_vector()
                    new_coord = eval(wy_coord)
                    specie = str(element)
                    coords_list.append(new_coord)
                    species_list.append(specie)
                    # test_struct.append(specie, new_coord)

                    dist_check = True  # assume its good until proven otherwise
                    fsg_test_struct = Structure.from_spacegroup(
                        spacegroup, struct_lattice, species_list, coords_list
                    )
                    #!!! speed improvements to be made here...?
                    # check distances in full unit cell to see if its too close to another site
                    fsg_dm = fsg_test_struct.distance_matrix

                    for x, specie1 in enumerate(new_composition):
                        specie1_indicies = fsg_test_struct.indices_from_symbol(
                            specie1.symbol
                        )
                        for y, specie2 in enumerate(new_composition):
                            specie2_indicies = fsg_test_struct.indices_from_symbol(
                                specie2.symbol
                            )
                            dist_cutoff = cutoff_matrix[x][y]
                            combos = itertools.product(
                                specie1_indicies, specie2_indicies
                            )
                            for s1, s2 in combos:
                                if s1 == s2:
                                    continue
                                distance = fsg_dm[s1][s2]
                                if distance < dist_cutoff:
                                    dist_check = False

                    # see if the distance check passed or not
                    if not dist_check:
                        # if not, remove the site and try again
                        # test_struct.pop() # do i have to pop from strucutre??
                        species_list.pop()
                        coords_list.pop()
                        continue

                # see if all checks (mult, avail, dist) above passed
                if dist_check:
                    # mult_check signifies whether total nsites has been reached so it is not necessarily true at this point
                    # still have to assign a coordinate to wy site even if mult_check isn't true
                    if mult_check:
                        master_check = (
                            True  # if mult_check is true flg end of main while loop
                        )
                    # if mult_check is false, another wy site will be assigned to the element
                    # if mult_check is true, program will move onto next element
                    continue
                else:
                    # if maxed out of attempts restart main while loop to pick different wy site
                    if element_sites:
                        element_sites.pop()
                        mults.pop()
                    # we want to restart the while loop
                    # since master_check was never changed, we dont need to reset it
                    # master_check=False

            # append wy letters for element to master list of wy letters
            master_wy_list.append(element_sites)

        return fsg_test_struct



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Eddrmm(ErrorHandler):
    """
    This is an error caused by failed call to LAPCK.

    This could be caused by a number of things, such as instability of the
    RMM-DIIS diagonalisation algorithm, an unreasonable crystal geometry, or
    even just incorrect installation of LAPCK.

    Further discussion is also located on VASP's forum
    [here](https://www.vasp.at/forum/viewtopic.php?f=3&t=214&p=215).
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["WARNING in EDDRMM: call to ZHEGV failed"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # RMM algorithm is not stable for this calculation
        if incar.get("ALGO", "Normal") in ["Fast", "VeryFast"]:
            incar["ALGO"] = "Normal"
            correction = "switched ALGO to Normal"
        else:
            # Halve the POTIM
            current_potim = incar.get("POTIM", 0.5)
            new_potim = current_potim / 2
            incar["POTIM"] = new_potim
            correction = f"switch POTIM from {current_potim} to {new_potim}"

        # Check the current ICHARG setting, where default is 0
        # If the ICHARG is less than 10, then we want to delete the CHGCAR
        # and WAVECAR to ensure the next run is a clean start.
        current_icharg = incar.get("ICHARG", 0)
        if current_icharg < 10:

            # check if IMAGES is in the INCAR. If so, we have a NEB calculation,
            # and the there will be a different organization of folders.
            if "IMAGES" in incar.keys():
                # here, folders will be 00, 01, 02, etc. with CHGCARs/WAVECARs
                # in each. We iterate through and delete all of them.
                subdirectories = [d for d in os.listdir(directory) if d.isnumeric()]
                for subdir in subdirectories:
                    chgcar_filename = os.path.join(directory, subdir, "CHGCAR")
                    wavecar_filename = os.path.join(directory, subdir, "WAVECAR")
                    if os.path.exists(chgcar_filename):
                        os.remove(chgcar_filename)
                    if os.path.exists(wavecar_filename):
                        os.remove(wavecar_filename)
                correction += " and deleted CHGCARs + WAVECARs for all images"

            # Otherwise, we have a normal VASP calculation
            else:
                os.remove(os.path.join(directory, "CHGCAR"))
                os.remove(os.path.join(directory, "WAVECAR"))
                correction += " and deleted CHGCAR + WAVECAR"
        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

import os

from pymatgen.io.vasp.outputs import Vasprun

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class IncorrectSmearing(ErrorHandler):
    """
    This checks if a calculation is a metal (zero bandgap), has been run with
    ISMEAR=-5, and is not a static calculation. This type of smearing is only
    appropriate for the relaxation of semiconductors and insulators -- not
    metals. If this occurs, this handler will rerun the calculation
    using the smearing settings appropriate for metals (ISMEAR=2, SIGMA=0.2).
    """

    is_monitor = False
    filename_to_check = "vasprun.xml"

    def check(self, directory: str) -> bool:

        # establish the full path to the output file
        filename = os.path.join(directory, self.filename_to_check)

        # we need this inside of a try/except clause because if the xml is
        # poorly formatted, then there is another issue at play. We only
        # want to check for the incorrect smearing here.
        try:
            # load the xml file and only parse the bare minimum
            xmlReader = Vasprun(
                filename=filename,
                parse_dos=False,
                parse_eigen=True,
                parse_projected_eigen=False,
                parse_potcar_file=False,
                exception_on_bad_xml=True,
            )

            # If all three of these conditions are met, then we have an error:
            #   (1) bandgap is zero
            #   (2) tetrahedron smearing was used (ISMEAR<-3)
            #   (3) relaxation is performed (NSW>1)
            if (
                xmlReader.eigenvalue_band_properties[0] == 0
                and xmlReader.incar.get("ISMEAR", 1) < -3
                and xmlReader.incar.get("NSW", 0) > 1
            ):
                return True

        except Exception:
            # any error above means there is some other problem, so we say
            # there is not an IncorrectSmearing error
            pass

        # if we reach this point, there is no error
        return False

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # update the smearing method to one appropriate for metals
        incar["ISMEAR"] = 2
        incar["SIGMA"] = 0.2
        correction = "switched ISMEAR to 2 amd SIGMA to 0.2"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

"""
This module creates all database tables for vasp relaxation workflows.

With the exception of StagedRelaxation, the data stored in each table is exactly 
the same -- so they all inherit from the `Relaxation` class without adding any
features. See `simmate.database.base_data_types.relaxation` for details.

StagedRelaxation on the other hand is a NestedCalculation -- in that it connects
a series of relaxation. See `simmate.database.base_data_types.calculation_nested`
for details.
"""

from simmate.database.base_data_types import Relaxation, NestedCalculation

# Between all of the different relaxations that simmate runs, there's no
# difference between any of the datatables we store results in. The difference
# is only HOW the relaxation was ran (i.e. the VASP settings used), which is
# why we store them in separate tables.

MITRelaxation, MITIonicStep = Relaxation.create_subclasses("MIT", module=__name__)

(
    MatprojRelaxation,
    MatprojIonicStep,
) = Relaxation.create_subclasses("Matproj", module=__name__)

(
    MatVirtualLabCINEBEndpointRelaxation,
    MatVirtualLabCINEBEndpointIonicStep,
) = Relaxation.create_subclasses("MatVirtualLabCINEBEndpoint", module=__name__)

Quality00Relaxation, Quality00IonicStep = Relaxation.create_subclasses(
    "Quality00",
    module=__name__,
)

Quality01Relaxation, Quality01IonicStep = Relaxation.create_subclasses(
    "Quality01",
    module=__name__,
)

Quality02Relaxation, Quality02IonicStep = Relaxation.create_subclasses(
    "Quality02",
    module=__name__,
)

Quality03Relaxation, Quality03IonicStep = Relaxation.create_subclasses(
    "Quality03",
    module=__name__,
)

Quality04Relaxation, Quality04IonicStep = Relaxation.create_subclasses(
    "Quality04",
    module=__name__,
)

StagedRelaxation = NestedCalculation.create_subclass_from_calcs(
    "StagedRelaxation",
    [
        Quality00Relaxation,
        Quality01Relaxation,
        Quality02Relaxation,
        Quality03Relaxation,
        Quality04Relaxation,
    ],
    module=__name__,
)



# -*- coding: utf-8 -*-

##############################################################################

"""

IMPORTANT!!!

We can't use the biblao server for our wyckoff data.

This is because a number of spacegroups are written based off of the wrong lattice.

For example, spacegroup 166 is written with the hexagonal lattice vectors, which
makes the smallest multiplicity 3 among all wy_sites. This in turn makes it so 
there is no possible wyckoff group combination that gives 2:1 stoich -- this is
misleading because we know that Ca2N has the spacegroup 166

"""

##############################################################################


from selenium import webdriver
import pandas as pd
from tqdm import tqdm

# launch the webbrowser
driver = webdriver.Chrome(
    executable_path="/snap/bin/chromium.chromedriver"
)  # /snap/bin/chromium gives errors

table_info = []
for spacegroup in tqdm(range(1, 231)):

    # load the webpage
    driver.get("https://www.cryst.ehu.es/cryst/get_wp.html")

    # find the input element (it should be empty to start, otherwise use inputElement.clear())
    inputElement = driver.find_element_by_name("gnum")

    # fill the input element with the spacegroup number
    inputElement.send_keys(spacegroup)

    # click the button that says "Standard/Defualt Settings"
    driver.find_element_by_name("standard").click()

    table = driver.find_elements_by_tag_name("table")[
        1
    ]  # grab all tables and then the one we want is the 2nd
    rows = table.find_elements_by_tag_name(
        "tr"
    )  # grab all the rows in that table. Note some rows have a table (w. more rows) inside so this isn't ideal.

    # first row is the table headers # THIS DOESNT NEED TO BE DONE EVERY TIME - consider moving for speed up
    headers = rows[0].text
    # remove the "\n" characters
    headers = headers.replace("\n", "")
    # split the string into a list at each ' '
    headers = headers.split(" ")
    # I'd like the first header to be the spacegroup
    headers = ["SpaceGroup", headers[0], headers[1], headers[2], headers[3]]

    # the second row is a spacer (empty)
    # then after that it's a mixture of the full row that we want with some duplicate "rows" (the last column)
    for row in rows[2:]:

        # grab the string
        info = row.text
        # remove the "\n" characters
        info = info.replace("\n", " ")
        # split the string into a list at each ' '
        info = info.split(" ")

        # this split() sometimes gives us the incorrect result because the entry in Sitesymmetry column might have a space in it
        # spacegroup 229 for example, wy_site 'g' has Sitesymmetry='mm2 . .' and wy_site 'd' has Sitesymmetry='-4m. 2'
        # to fix this, I look at at the indicies immediately following the typical Sitesymmetry position (at index=2)
        # and if that index has an entry length smaller than 3, I deem it part of the Sitesymmetry.
        # This works because (x,y,z) will have length >= 7
        # This code is pretty ugly... But we are just webscraping and formmatting data the way we want, so oh well!
        info_fixed = []
        for i, entry in enumerate(info):
            # first two entries of "Multiplicity" and "Wyckoffletter" are fine - just append them
            if i == 0 or i == 1:
                info_fixed.append(entry)
            # index 2 is the start of the "Sitesymmetry" entry
            elif i == 2:
                # look at the next 2 entries (this issue never goes past two spaces in the Sitesymmetry label)
                # and evaluate whether or not they should be apart of the Sitesymmetry label (see logic explained above)
                # sometimes there is no index at info[i+1], so check the len first!
                if len(info) > 3:
                    next1 = info[i + 1]
                    if len(next1) < 3:
                        entry += " " + next1
                # sometimes there is no index at info[i+2], so check the len first!
                if len(info) > 4:
                    next2 = info[i + 2]
                    if len(next2) < 3:
                        entry += " " + next2
                # entry should now be the correct label for SiteSymmetry so append it
                info_fixed.append(entry)
            # for these two indecies, only append if they are not a part of the SiteSymmetry label
            elif i == 3 or i == 4:
                if len(entry) > 3:
                    info_fixed.append(entry)
            # the rest of the entries we can safely assume are coordinates
            else:
                info_fixed.append(entry)
        # take these fixes and override the original info
        info = info_fixed

        # we need to check if this is actually a row we want.
        # the logic I use here is that the multiplicity won't ever be more than 3 digits (max is 192)
        # while the an (x,y,z) entry will alway have more than 3
        # so therefore the first column should be =< 3 if it's a row we want
        if len(info[0]) <= 3:
            # the first 3 entries match the first 3 headers of the table
            # the rest should go under the 4th header. Let's move them all in to one entry.
            info = [spacegroup, int(info[0]), info[1], info[2], info[3:]]

            # reformat the '(x,y,z)' strings in info[4] list to ['x','y','z']
            # info[4] = [n[1:-1].split(',') for n in info[4]]

            # add the result to the dataframe
            table_info.append(info)

# close the window
driver.close()

# make DataFrame
df = pd.DataFrame(data=table_info, columns=headers)

# export to csv file
df.to_csv("wyckoffdata.csv", index=False)

##############################################################################

# set this if you want to run headless (i.e. in the background)
# options = webdriver.ChromeOptions()
# options.headless = True
# and then add the setting to this call: webdriver.Chrome(options = options)


# customize the window if you'd like
# driver.minimize_window()
# driver.set_window_size(width=1920, height=1080)

##############################################################################



# -*- coding: utf-8 -*-

"""
This file establishes the info needed for installing your project via pip. This
then let's you import your code! For example, you can do...

``` python
from example_app.apps import ExampleAppConfig
```
    
As your project grows, you may want to update the requirements for it or even
upload it to PyPI, which let's other people download and install your code. 
To see how this setup.py file can grow, you can take a look at other example files:
    Simmate's setup.py is at... https://github.com/jacksund/simmate/blob/main/setup.py
    Another example one is at... https://github.com/pypa/sampleproject

"""

from setuptools import setup, find_packages

setup(
    name="{{ project_name }}",
    version="0.0.0",
    python_requires=">=3.10, <4",
    packages=find_packages(where="."),
    install_requires=["simmate"],
)



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class Brions(ErrorHandler):
    """
    This fixes an internal VASP error by increasing POTIM.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["BRIONS problems: POTIM should be increased"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # make the fix
        current_potim = incar.get("POTIM", 0.5)
        new_potim = current_potim + 0.1
        incar["POTIM"] = new_potim
        correction = f"switched POTIM from {current_potim} to {new_potim}"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

import os

from simmate.conftest import copy_test_files, make_dummy_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import Zpotrf


def test_eddrmm(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="zpotrf",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")
    chgcar_filename = os.path.join(tmpdir, "CHGCAR")
    wavecar_filename = os.path.join(tmpdir, "WAVECAR")
    oszicar_filename = os.path.join(tmpdir, "OSZICAR")

    # init class with default settings
    error_handler = Zpotrf()

    # Confirm an error IS NOT found
    error_handler.filename_to_check = "vasp.no_error"
    assert error_handler.check(tmpdir) == False

    # Confirm an error IS found
    error_handler.filename_to_check = "vasp.out"
    assert error_handler.check(tmpdir) == True

    # Attempt to fix the error when we have >1 valid ionic steps in the OSZICAR
    make_dummy_files(chgcar_filename, wavecar_filename)
    fix = error_handler.correct(tmpdir)
    assert (
        fix
        == "set ISYM to 0 and switched POTIM from 0.5 to 0.25 and deleted CHGCAR and WAVECAR"
    )
    incar = Incar.from_file(incar_filename)
    assert incar["ISYM"] == 0
    assert incar["POTIM"] == 0.25
    assert not os.path.exists(chgcar_filename)
    assert not os.path.exists(wavecar_filename)

    # remove OSZICAR to the remaining checks/fixes
    os.remove(oszicar_filename)

    # Make first attempt at fixing the error
    make_dummy_files(chgcar_filename, wavecar_filename)
    fix = error_handler.correct(tmpdir)
    assert fix == "set ISYM to 0 and deleted CHGCAR and WAVECAR"
    incar = Incar.from_file(incar_filename)
    assert Incar.from_file(incar_filename)["ISYM"] == 0
    assert not os.path.exists(chgcar_filename)
    assert not os.path.exists(wavecar_filename)

    # Make second attempt at fixing the error
    with open(incar_filename, "w") as file:
        file.write("NSW = 1 \nISIF = 2")
    make_dummy_files(chgcar_filename, wavecar_filename)
    fix = error_handler.correct(tmpdir)
    assert fix == "scaled the structure lattice by +20% and deleted CHGCAR and WAVECAR"



# -*- coding: utf-8 -*-


class ASEStructure:

    # see source: https://gitlab.com/ase/ase/-/blob/master/ase/ga/startgenerator.py
    # see tutorials: https://wiki.fysik.dtu.dk/ase/ase/ga/ga.html#module-ase.ga

    def __init__(
        self,
        composition,  # must be a pymatgen Composition object
        ratio_of_covalent_radii=0.5,
    ):

        # we can assume the user has ASE installed because it is a dependency of PyMatgen
        #!!! it looks like the ase.ga module is actively changing so version may introduce errors
        from ase.ga.startgenerator import StartGenerator
        from ase.ga.utilities import closest_distances_generator

        # the slab variable is there to specify dimensionality. The code really just uses the pbc setting
        #!!! I need to doublecheck this
        from ase import Atoms

        slab = Atoms(
            pbc=True
        )  #!!! I assume 3D structure for now. I could do things like pbc=[1,1,0] for 2D in the future though

        # the blocks input is just a list of building blocks and how many to include of each
        # for example, Ca2N would have [('Ca', 2), ('N', 1)].
        # let's convert the pymatgen composition object to this format - other formats are allowed too (see docs)
        blocks = [
            (element.symbol, int(composition[element])) for element in composition
        ]

        # the closest_distances_generator is exactly the same as an element-dependent distance matrix
        # expect ASE puts this in dictionary form
        # the function requires a list of element integers
        element_ints = [element.number for element in composition]
        # the default of the ratio of covalent radii (0.5) is based on the ASE tutorial of this function
        element_distance_matrix = closest_distances_generator(
            element_ints, ratio_of_covalent_radii
        )

        # boundry limits on the lattice
        # I only do angle limits for now because I try to set the volume below
        from ase.ga.utilities import CellBounds

        cellbounds = CellBounds(
            bounds={
                "phi": [60, 120],
                "chi": [60, 120],
                "psi": [60, 120],
            }
        )
        #'a': [3, 50], 'b': [3, 50], 'c': [3, 50]})

        # predict volume
        volume_guess = composition.volume_estimate()

        self.ase = StartGenerator(
            slab=slab,
            blocks=blocks,
            blmin=element_distance_matrix,
            number_of_variable_cell_vectors=3,  #!!! I understand this as number of unique vectors. Is that right?
            # box_to_place_in=None, # use the entire unitcell
            box_volume=volume_guess,  # target this volume when making the lattice
            # splits=None, # increases translational sym (good for large atom counts - forces symmetry)
            cellbounds=cellbounds,
            # test_dist_to_slab=True, # ensure distances arent too close
            # test_too_far=True, # ensure no atom is isolated (distance is >2x the min cutoff)
            # rng=np.random
        )

    def new_structure(self, spacegroup=None, lattice=None, species=None, coords=None):

        # if the user supplied a sites, it's going to be overwritten.
        if coords or species:
            #!!! this error will always show up in my current form of sample.py -- should I change that?
            #!!! the only time the error doesn't show up is when no validators fail on the object
            print(
                "Warning: ASE.ga does not allow for specification of atomic sites and your "
                "specification here will be overwritten."
            )

        # if the user supplied a lattice, it's going to be overwritten.
        if lattice:
            print(
                "Warning: While ASE.ga allows for specifying a fixed lattice, "
                "our wrapper does not currently support this. Contact our team if you"
                " would like this feature incorporated."
            )

        # if a spacegroup is not specified, grab a random one from our options
        # no check is done to see if the spacegroup specified is compatible with the vector_generator built
        if spacegroup:
            print(
                "Warning: ASE.ga does not use symmetry to generate their structures "
                "and specifying this has no effect."
            )

        # now make the new structure using ase,ga.startgenerator
        # NOTE: all of these options are set in the init except for the spacegroup
        structure_ase = self.ase.get_new_candidate(
            maxiter=1000
        )  #!!! I should set a limit on the iterations here

        # if creation fails, None is returned
        if not structure_ase:
            return False

        # convert the ase Atoms object to pymatgen Structure
        from pymatgen.io.ase import AseAtomsAdaptor

        structure = AseAtomsAdaptor.get_structure(structure_ase)

        return structure



# -*- coding: utf-8 -*-

from simmate.command_line.base_command import simmate


def test_base_command(command_line_runner):
    result = command_line_runner.invoke(simmate)
    assert result.exit_code == 0



# -*- coding: utf-8 -*-

"""
Build a workflow in Simmate involves piecing together our Tasks and Database 
tables. So before you start editting this file, make sure you have gone through
and editted the `models.py` and `tasks.py` files.

This is the most basic Workflow that you may build. For complex workflows that
require several calculations, other python logic, and parallelization, make
sure you read through the documentation for 
[the Workflow class](https://jacksund.github.io/simmate/simmate/workflow_engine/workflow.html)
"""

# To access advanced functionality, we should always inherit from the base Workflow
from simmate.workflow_engine import Workflow

# Import our tables and tasks from the other files.
# Note, the format `from <place> import <name> as <newname>` simply renames the class so
# that the table/task classes don't have the same name in this file.
from .models import ExampleRelaxation as ExampleRelaxationTable
from .tasks import ExampleRelaxation as ExampleRelaxationTask

# Now build our workflow
class Relaxation__Vasp__MyCustomSettings(Workflow):
    # Simply set these two variables to your Task and DatabaseTable classes!
    s3task = ExampleRelaxationTask
    database_table = ExampleRelaxationTable



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.relaxation import (
    Quality00Relaxation as Quality00RelaxationTask,
)
from simmate.calculators.vasp.database.relaxation import (
    Quality00Relaxation as Quality00RelaxationResults,
)


class Relaxation__Vasp__Quality00(Workflow):
    s3task = Quality00RelaxationTask
    database_table = Quality00RelaxationResults
    description_doc_short = "bare minimum settings meant for highly unreasonable structures. lattice volume kept fixed"



# -*- coding: utf-8 -*-

from simmate.workflow_engine import S3Task


# TODO: The chgsum.pl script will be replaced with a simple python function
# that just sums the two files. It might not be as fast but it removes one
# executable file from having to be in the user's path. So in the future, this
# Task will be depreciated/removed into the BaderAnalysis.setup method.
class CombineCHGCARs(S3Task):
    """
    This tasks simply sums two charge density files into a new file. It uses
    a script from the Henkleman group.
    """

    command = "chgsum.pl AECCAR0 AECCAR2 > chgsum.out"
    monitor = False
    required_files = ["CHGCAR", "AECCAR0", "AECCAR2"]



# -*- coding: utf-8 -*-

from simmate.toolkit.creators.vector import UniformlyDistributedVectors


#!! NOTE: I should make a copy of this that is RandomSymSites, where I limit the
# random sites generated to the asym unit of the specified spacegroup
class RandomSites:
    def __init__(
        self,
        composition,  # must be a pymatgen Composition object
        coords_generation_method=UniformlyDistributedVectors,
        coords_gen_options=dict(),
    ):

        # save the composition for reference
        self.composition = composition

        # establish the coords generator
        self.coords_generator = coords_generation_method(**coords_gen_options)

    def new_sites(self):

        # make an empty list to store all the fractional coordinates
        coords_list = []  # regular list.append is faster than numpy.append
        # as well as an empty list for the species
        species_list = []

        # let's go through each element in composition one at a time
        for element in self.composition:
            # element is a key to composition that returns nsites
            # for example, composition['Ca'] = 2 for Ca2N
            nsites = int(self.composition[element])
            for site in range(nsites):
                # use the generator to create random fractional coordinates
                coords = self.coords_generator.new_vector()
                # append the coords and species to our output lists
                coords_list.append(coords)
                species_list.append(element)

        # maybe make these into pymatgen Site objects? I don't see any advantage to Site object,
        # because we will later need PeriodicSite objects and there's no direct conversion method
        return species_list, coords_list



# -*- coding: utf-8 -*-

"""
This module defines the Structure class.

It is a very basic extension PyMatGen's core Structure class, as it only adds
a few extra methods and does not change any other usage.
"""

import os

from pymatgen.core import Structure as PymatgenStructure


class Structure(PymatgenStructure):
    # Leave docstring blank and just inherit from pymatgen

    database_object = None  # simmate.database.base_data_types.Structure
    """
    If this structure came from a `simmate.database.base_data_types.Structure`
    object, then this attribute will be set to the original database object.
    
    Otherwise, this will be left as `None`.
    """
    # Note, we don't set the type here because it is a cross-module dependency
    # that is frequently not required

    def get_sanitized_structure(self):
        """
        Run symmetry analysis and "sanitization" on the pymatgen structure
        """

        # Make sure we have the primitive unitcell first
        # We choose to use SpagegroupAnalyzer (which uses spglib) rather than pymatgen's
        # built-in Structure.get_primitive_structure function:
        #   structure = structure.get_primitive_structure(0.1) # Default tol is 0.25

        # Default tol is 0.01, but we use a looser 0.1 Angstroms
        # from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
        # structure_primitive = SpacegroupAnalyzer(structure, 0.1).find_primitive()
        # BUG: with some COD structures, this symm analysis doesn't work. I need
        # to look into this more and figure out why.

        # Default tol is 0.25 Angstroms
        structure_primitive = self.get_primitive_structure()

        # Convert the structure to a "sanitized" version.
        # This includes...
        #   (i) an LLL reduction
        #   (ii) transforming all coords to within the unitcell
        #   (iii) sorting elements by electronegativity
        structure_sanitized = structure_primitive.copy(sanitize=True)

        # return back the sanitized structure
        return structure_sanitized

    @classmethod
    def from_dynamic(cls, structure):
        """
        This is an experimental feature.

        Possible structure formats include...
            object of toolkit structure
            dictionary of toolkit structure
            dictionary of...
                (1) python path to calculation datatable
                (2) one of the following (only one is used in this priority order):
                    (a) prefect flow id
                    (b) calculation id
                    (c) directory
                    ** these three are chosen because all three are unique for every
                    single calculation and we have access to different ones at different
                    times!
                (3) (optional) attribute to use on table (e.g. structure_final)
                    By default, we assume calculation table is also a structure table
            filename for a structure (cif, poscar, etc.) [TODO]
        """

        # keep track of this flag too, which we use in the next sections on loading
        # the directory and source
        is_from_past_calc = False

        # if the input is already a pymatgen structure, just return it back
        if isinstance(structure, PymatgenStructure):
            structure_cleaned = structure

        # if the "@module" key is in the dictionary, then we have a pymatgen
        # structure dict which we convert to a pymatgen object and return
        elif isinstance(structure, dict) and "@module" in structure.keys():
            structure_cleaned = cls.from_dict(structure)

        # if there is a database_table key, then we are pointing to the simmate
        # database for the input structure
        elif isinstance(structure, dict) and "database_table" in structure.keys():
            is_from_past_calc = True
            structure_cleaned = cls.from_database_dict(structure)

        # if the value is a str and it relates to a filepath, then we load the
        # structure from a file.
        elif isinstance(structure, str) and os.path.exists(structure):
            structure_cleaned = cls.from_file(structure)

        # Otherwise an incorrect format was given
        else:
            raise Exception(
                "Unknown format provided for structure input. "
                f"{type(structure)} was provided. If you are trying "
                "to provide a filepath (str), make sure you don't have "
                "any typos and that the path is relative to the working "
                "directory."
            )

        # add this attribute to help with error checking in other methods
        structure_cleaned.is_from_past_calc = is_from_past_calc

        return structure_cleaned

    # -------------------------------------------------------------------------

    # All methods below wrap converters from the `simmate.file_converters.structure`
    # module. To keep this class modular (from the database and optional deps),
    # imports are done lazily. This means we import the relevant convert once
    # the method is called -- rather than when this module is initially loaded.

    @classmethod
    def from_database_dict(cls, structure: dict):
        from simmate.file_converters.structure.database import DatabaseAdapter

        return DatabaseAdapter.get_toolkit_from_database_dict(structure)

    @classmethod
    def from_database_object(cls, structure: dict):
        from simmate.file_converters.structure.database import DatabaseAdapter

        return DatabaseAdapter.get_toolkit_from_database_object(structure)

    @classmethod
    def from_database_string(cls, structure_string: str):
        from simmate.file_converters.structure.database import DatabaseAdapter

        return DatabaseAdapter.get_toolkit_from_database_string(structure_string)

    # TODO: from_cif, from_poscar, from_ase, from_jarvis, etc.
    # TODO: to_cif, to_poscar, to_ase, to_jarvis, etc.



# -*- coding: utf-8 -*-

"""
This defines the command for running a test-server locally.
"""

import click


@click.command()
def run_server():
    """
    This runs a website test server locally for Simmate. You can then view the
    working website at http://127.0.0.1:8000/

    This command is exactly the same as running...

    django runserver --settings=simmate.configuration.django.settings
    """

    from simmate.database import connect
    from django.core.management import call_command

    call_command("runserver")


# explicitly list functions so that pdoc doesn't skip them
__all__ = ["run_server"]



# -*- coding: utf-8 -*-

from numpy.random import randint, choice

from pymatgen.analysis.structure_matcher import StructureMatcher

from simmate.toolkit.transformations.base import Transformation


class AtomicPermutation(Transformation):

    # Two atoms of different types are exchanged a variable number of times
    # See: https://uspex-team.org/static/file/JChemPhys-USPEX-2006.pdf

    io_scale = "one_to_one"
    ninput = 1
    use_multiprocessing = False

    #!!! should I add an option for distribution of exchange number? Potentially use creation.vector objects?
    # USPEX states that they do this - and select from a normal distribution
    def __init__(self, min_exchanges=1, max_exchanges=5):

        # save inputs for reference
        self.min_exchanges = min_exchanges
        self.max_exchanges = max_exchanges

        #!!! when doing multiple exchanges, I do not check to see if an exchange has been done before - I can change this though
        # Therefore one exchange could undo another and we could be back where we started
        # There are also scenarios where all sites are equivalent and atomic permutation cannot create a new structure
        # An example of this is NaCl, where an exchange still yields an identical structure
        # This leaves a chance that we end up with an identical structure to what we started with.
        # Therefore, I must hava a structurematcher object that I use to ensure we have a new structure
        # We try making a new structure X number of times (see max_attempts above) and if we can't, we failed to mutate the structure!
        self.structure_matcher = (
            StructureMatcher()
        )  #!!! I should lower the tolerances here!

    def apply_transformation(self, structure, max_attempts=100):

        # grab a list of the elements
        #!!! consider moving this to init to save time at the cost of more in memory
        elements = structure.composition.elements

        # This mutation is not possible for structures that have only one element
        if len(elements) == 1:
            print(
                "You cannot perform an atomic permutation on structure that only has one element type!!"
            )
            return False
        #!!! add another elif for when all sites are equivalent and atomic permutation cannot create a new structure

        for attempt in range(max_attempts):  # I could alternatively do a while loop

            # make a deepcopy of the structure so that we aren't modifying it inplace
            # this also allows us to compare the new structure to the original
            # placing this at the top of the loop also resets the structure for us each time
            new_structure = structure.copy()

            # grab a random integer within the exchange range (min/max) defined above
            nexchanges = randint(low=self.min_exchanges, high=self.max_exchanges)

            # perform an exchange of two random atom types X number of times
            for n in range(nexchanges):
                # grab two elements of different types
                element1, element2 = choice(elements, size=2)

                # select a random index of element1 and element2
                index1 = choice(new_structure.indices_from_symbol(element1.symbol))
                index2 = choice(new_structure.indices_from_symbol(element2.symbol))

                # now exchange the species type of those two sites
                new_structure.replace(index1, element2)
                new_structure.replace(index2, element1)

            # see if the new structure is different from the original!
            # check will be True if the new structure is the same as the original
            check = self.structure_matcher.fit(structure, new_structure)
            if not check:
                # we successfully make a new structure!
                return new_structure
            # else continue

        # if we make it this far, then we hit our max_attempts limit without making a new structure
        # therefore, we failed the mutation
        print("Failed to make a new structure that is different from the original")
        return False



# -*- coding: utf-8 -*-

"""
Runs a NEB on all unique pathways within a structure. 

The folder tree looks like...
```
simmate-task-12345/  # determined by simmate.utilities.get_directory
    ├── bulk_relaxation
    ├── bulk_static_energy
    ├── migration_hop_00
    ├── migration_hop_01
    ...
    └── migration_hop_N  # all migration_hop folders have the same structure
        ├── endpoint_relaxation_start
        ├── endpoint_relaxation_end
        ├── 01
        ├── 02
        ├── 03
        ...
        └── N  # corresponds to image number
```
"""

import os
from typing import List

from simmate.toolkit import Structure
from simmate.toolkit.diffusion import DistinctPathFinder, MigrationHop
from simmate.workflow_engine import Workflow, task
from simmate.workflow_engine.common_tasks import (
    load_input_and_register,
    parse_multi_command,
)
from simmate.calculators.vasp.workflows.relaxation.mit import (
    Relaxation__Vasp__Mit,
)
from simmate.calculators.vasp.workflows.static_energy.mit import (
    StaticEnergy__Vasp__Mit,
)
from simmate.calculators.vasp.workflows.diffusion.neb_single_path import (
    Diffusion__Vasp__NebSinglePath,
)
from simmate.calculators.vasp.database.nudged_elastic_band import (
    MITDiffusionAnalysis,
)
from simmate.database.base_data_types.nudged_elastic_band import (
    DiffusionAnalysis as DiffusionAnalysisTable,
)


class Diffusion__Vasp__NebAllPaths(Workflow):
    """
    Runs a full diffusion analysis on a bulk crystal structure using NEB.

    The bulk structure will be geometry optimized and then
    `simmate.toolkit.diffusion.DistinctPathFinder` is used to find all
    symmetrically unique migration hops in the structure up until the hops
    become percolating (>0-D). For each unique hop, the workflow
    diffusion/single_path is submitted.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation/mit
        - static-energy/mit
        - a mini task that identifies unique migration hops
        - (for each hop) diffusion/single-path

    If you are running this workflow via the command-line, you can run this
    with...

    ``` bash
    simmate workflows run diffusion/all-paths -s example.cif -c "cmd1; cmd2; cmd3"
    ```

    Note, the `-c` here is very important! Here we are passing three commands
    separated by semicolons. Each command is passed to a specific workflow call:

        - cmd1 --> used for bulk crystal relaxation and static energy
        - cmd2 --> used for endpoint supercell relaxations
        - cmd3 --> used for NEB

    Thus, you can scale your resources for each step. Here's a full -c option:

    -c "vasp_std > vasp.out; mpirun -n 12 vasp_std > vasp.out; mpirun -n 70 vasp_std > vasp.out"
    """

    database_table = MITDiffusionAnalysis

    description_doc_short = "runs NEB for all symmetrically unique paths"

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        migrating_specie: str,
        command: str = None,
        source: dict = None,
        directory: str = None,
        is_restart: bool = False,
    ):
        # command list expects three subcommands:
        #   command_bulk, command_supercell, and command_neb
        #
        # I separate these out because each calculation is a very different scale.
        # For example, you may want to run the bulk relaxation on 10 cores, the
        # supercell on 50, and the NEB on 200. Even though more cores are available,
        # running smaller calculation on more cores could slow down the calc.
        subcommands = parse_multi_command(
            command,
            commands_out=["command_bulk", "command_supercell", "command_neb"],
        )

        # Load our input and make a base directory for all other workflows to run
        # within for us.
        parameters_cleaned = load_input_and_register(
            structure=structure,
            source=source,
            directory=directory,
            command=command,
            migrating_specie=migrating_specie,
            is_restart=is_restart,
            register_run=False,
        )

        # Our step is to run a relaxation on the bulk structure and it uses our inputs
        # directly. The remaining one tasks pass on results.
        bulk_relax_result = Relaxation__Vasp__Mit.run(
            structure=parameters_cleaned["structure"],
            command=subcommands["command_bulk"],
            directory=parameters_cleaned["directory"]
            + os.path.sep
            + Relaxation__Vasp__Mit.name_full,
            is_restart=is_restart,
        ).result()

        # A static energy calculation on the relaxed structure. This isn't necessarily
        # required for NEB, but it takes very little time.
        bulk_static_energy_result = StaticEnergy__Vasp__Mit.run(
            structure={
                "database_table": Relaxation__Vasp__Mit.database_table.__name__,
                "directory": bulk_relax_result["directory"],
                "structure_field": "structure_final",
            },
            command=subcommands["command_bulk"],
            directory=parameters_cleaned["directory"]
            + os.path.sep
            + StaticEnergy__Vasp__Mit.name_full,
            is_restart=is_restart,
        ).result()

        # This step does NOT run any calculation, but instead, identifies all
        # diffusion pathways and builds the necessary database entries.
        migration_hop_ids = build_diffusion_analysis(
            diffusion_analyis_table=cls.database_table,
            structure={
                "database_table": StaticEnergy__Vasp__Mit.database_table.__name__,
                "directory": bulk_static_energy_result["directory"],
            },
            migrating_specie=migrating_specie,
            directory=parameters_cleaned["directory"],
            vacancy_mode=True,  # assumed for now
        )

        # Run NEB single_path workflow for all these.
        for i, hop_id in enumerate(migration_hop_ids):
            state = Diffusion__Vasp__NebSinglePath.run(
                migration_hop={
                    "migration_hop_table": "MITMigrationHop",
                    "migration_hop_id": hop_id,
                },
                directory=parameters_cleaned["directory"]
                + os.path.sep
                + f"{Diffusion__Vasp__NebSinglePath.name_full}.{str(i).zfill(2)}",
                diffusion_analysis_id=None,
                migration_hop_id=None,
                command=subcommands["command_supercell"]
                + ";"
                + subcommands["command_neb"],
                is_restart=is_restart,
            )  # we don't want to wait on results to in order to allow parallel runs


@task
def build_diffusion_analysis(
    diffusion_analyis_table: DiffusionAnalysisTable,
    structure: Structure,
    migrating_specie: str,
    vacancy_mode: bool,
    directory: str = "",
    **kwargs,
) -> List[MigrationHop]:
    """
    Given a bulk crystal structure, returns all symmetrically unique pathways
    for the migrating specie (up until the path is percolating). This
    also create all relevent database entries for this struture and its
    migration hops.

    #### Parameters

    - `structure`:
        bulk crystal structure to be analyzed. Can be in any format supported
        by Structure.from_dynamic method.

    - `migrating_specie`:
        Element or ion symbol of the diffusion specie (e.g. "Li")

    - `directory`:
        where to write the CIF file visualizing all migration hops. If no
        directory is provided, it will be written in the working directory.

    - `**kwargs`:
        Any parameter normally accepted by DistinctPathFinder
    """

    ###### STEP 1: creating the toolkit objects and writing them to file

    structure_cleaned = Structure.from_dynamic(structure)

    pathfinder = DistinctPathFinder(
        structure_cleaned,
        migrating_specie,
        **kwargs,
    )
    migration_hops = pathfinder.get_paths()

    # We write all the path files so users can visualized them if needed
    filename = os.path.join(directory, "migration_hop_all.cif")
    pathfinder.write_all_paths(filename, nimages=10)
    for i, migration_hop in enumerate(migration_hops):
        number = str(i).zfill(2)  # converts numbers like 2 to "02"
        # the files names here will be like "migration_hop_02.cif"
        migration_hop.write_path(
            os.path.join(directory, f"migration_hop_{number}.cif"),
            nimages=10,
        )

    ###### STEP 2: creating the database objects and saving them to the db

    # Create the main DiffusionAnalysis object that others will link to.
    da_obj = diffusion_analyis_table.from_toolkit(
        structure=structure_cleaned,
        migrating_specie=migrating_specie,
        vacancy_mode=vacancy_mode,
    )
    da_obj.save()
    # TODO: should I search for a matching bulk structure before deciding
    # to create a new DiffusionAnalysis entry?

    # grab the linked MigrationHop class
    hop_class = da_obj.migration_hops.model

    # Now iterate through the hops and add them to the database
    hop_ids = []
    for i, hop in enumerate(migration_hops):
        hop_db = hop_class.from_toolkit(
            migration_hop=hop,
            number=i,
            diffusion_analysis_id=da_obj.id,
        )
        hop_db.save()
        hop_ids.append(hop_db.id)

    # TODO: still figuring out if toolkit vs. db objects should be returned.
    # Maybe add ids to the toolkit objects? Or dynamic DB dictionaries?
    # For now I return the MigrationHop ids -- because this let's me
    # indicate which MigrationHops should be updated later on.
    return hop_ids



# -*- coding: utf-8 -*-


class LrfCommutatorHandler(ErrorHandler):
    """
    Corrects LRF_COMMUTATOR errors by setting LPEAD=True if not already set.
    Note that switching LPEAD=T can slightly change results versus the
    default due to numerical evaluation of derivatives.
    """

    is_monitor = True

    error_msgs = {"lrf_comm": ["LRF_COMMUTATOR internal error"]}

    def __init__(self, output_filename="std_err.txt"):
        """
        Initializes the handler with the output file to check.
        Args:
            output_filename (str): This is the file where the stderr for vasp
                is being redirected. The error messages that are checked are
                present in the stderr. Defaults to "std_err.txt", which is the
                default redirect used by :class:`custodian.vasp.jobs.VaspJob`.
        """
        self.output_filename = output_filename
        self.errors = set()
        self.error_count = Counter()

    def check(self):
        """
        Check for error.
        """
        self.errors = set()
        with open(self.output_filename, "r") as f:
            for line in f:
                l = line.strip()
                for err, msgs in LrfCommutatorHandler.error_msgs.items():
                    for msg in msgs:
                        if l.find(msg) != -1:
                            self.errors.add(err)
        return len(self.errors) > 0

    def correct(self):
        """
        Perform corrections.
        """
        backup(VASP_BACKUP_FILES | {self.output_filename})
        actions = []
        vi = VaspInput.from_directory(".")

        if "lrf_comm" in self.errors:
            if Outcar(zpath(os.path.join(os.getcwd(), "OUTCAR"))).is_stopped is False:
                if not vi["INCAR"].get("LPEAD"):
                    actions.append(
                        {"dict": "INCAR", "action": {"_set": {"LPEAD": True}}}
                    )

        VaspModder(vi=vi).apply_actions(actions)
        return {"errors": list(self.errors), "actions": actions}



# -*- coding: utf-8 -*-

from abc import ABC, abstractmethod


class Transformation(ABC):

    """
    This is an abstract base class (ABC) that shows you how to write your
    own Transformation class. It defines the requirements for you and raises
    errors if you forgot to define something.

    Using this abstract class, here's an example subclass that you could write
    -------------------------------------
    class SubClass(Transformation):
        io_type = 'one_to_one'
        ninput = 1
        use_multiprocessing = False
        def apply_transformation(self, structure):
            # do some transform on the input
            return new_structure
    -------------------------------------
    # example use
    t = SubClass()
    new_structure = t.apply_transformation(structure)
    -------------------------------------

    This is an updated version of pymatgen's transformation class because they
    do not support many_to_one or many_to_many:
    pymatgen.transformations.transformation_abc
    see https://pymatgen.org/pymatgen.transformations.transformation_abc.html
    """

    @property
    # @abstractmethod #!!! uncomment when I no longer assume the value
    def io_type(self):
        """
        Should I do StructureTransformation, LatticeTransformation, and
        SiteTransformation subclasses? Here, io_type could be either
        'structure', 'lattice', or 'site' as to indicate the input.
        """
        #!!! For now, I have all transformations as StructureTransformation
        #!!! so I assume io_type here. This may change in the future.
        return "structure"

    @property
    @abstractmethod
    def io_scale(self):
        """
        The input/output type for the class.
        This should be one of the following choices:
            - one_to_one
            - one_to_many
            - many_to_one
            - many_to_many # I have no examples of this yet
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def use_multiprocessing(self):
        """
        Whether or not this transformation can be done in parallel.
        Simply set to True or False.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def ninput(self):
        """
        Number of inputs required. For example, some transformations require
        two structures to be input. If one_to_* we know ninput = 1.
        """
        raise NotImplementedError

    @abstractmethod
    def apply_transformation(self):
        """
        The code that carries out the
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

import os
import yaml

from simmate.workflow_engine import Workflow
from simmate.utilities import copy_directory
from simmate.workflows.utilities import get_workflow


class Restart__Simmate__Automatic(Workflow):
    """
    Given a directory of a incomplete Simmate workflow run, this will use the
    metadata file to restart the calculation.

    Example yaml file:

    ``` yaml
    workflow-name: restart.simmate.automatic
    directory_old: path/to/calc
    directory_new: path/to/newcalc (optional)
    ```

    No other inputs are allowed bc the simmate_metadata.yaml will be used.
    """

    @classmethod
    def run_config(cls, directory_old: str, directory_new: str = None):

        # First copy over the directory
        directory_new_cleaned = copy_directory(directory_old, directory_new)

        # grab the metadata file in the new dir
        metadata_filename = os.path.join(
            directory_new_cleaned,
            "simmate_metadata.yaml",
        )
        with open(metadata_filename) as file:
            metadata = yaml.full_load(file)

        # Update the parameters from the metadata below
        input_parameters = metadata.copy()

        # remove settings that will be reset elsewhere
        input_parameters.pop("prefect_flow_run_id", None)
        input_parameters.pop("directory", None)
        input_parameters.pop("copy_previous_directory", None)
        input_parameters.pop("is_restart", None)

        # grab the workflow we need to run
        workflow_name = input_parameters.pop("workflow_name")
        workflow = get_workflow(workflow_name)

        # update the source
        input_parameters.pop("source", None)
        input_parameters["source"] = {
            "database_table": workflow.database_table.__name__,
            "directory": directory_old,
            # BUG: I should force directory_old to be an absolute path in order
            # to store it here, but I'm unsure how to properly handle different
            # file systems...
            "is_restart": True,
        }

        # now instead of calling the workflow's run method, we use its run_restart
        state = workflow.run(
            directory=directory_new_cleaned,
            is_restart=True,
            **input_parameters,
        )
        result = state.result()

        return result



# -*- coding: utf-8 -*-

# I store all of my models elsewhere, so this file simply exists to show django where
# they are located at. I do this based on the directions given by:
# https://docs.djangoproject.com/en/3.1/topics/db/models/#organizing-models-in-a-package

from simmate.database.base_data_types import Spacegroup



# -*- coding: utf-8 -*-

from simmate.utilities import get_doc_from_readme

__doc__ = get_doc_from_readme(__file__)

from .aflow import AflowStructure, AflowPrototype
from .cod import CodStructure
from .jarvis import JarvisStructure
from .materials_project import MatprojStructure
from .oqmd import OqmdStructure
from .utilities import load_remote_archives, load_default_sqlite3_build



# -*- coding: utf-8 -*-

import os
from pathlib import Path


def load_cluster_settings() -> dict:
    """
    Load our default cluster type and worker settings. This is loaded
    from the file `dask_cluster.yaml`.

    Note, module must be imported BEFORE we import any dask modules.
    """

    SIMMATE_DIRECTORY = os.path.join(Path.home(), "simmate")
    CLUSTER_YAML = os.path.join(SIMMATE_DIRECTORY, "dask_cluster.yaml")
    if os.path.exists(CLUSTER_YAML):
        os.environ.setdefault(
            "DASK_CONFIG",
            os.path.join(os.path.expanduser("~"), "simmate", "dask_cluster.yaml"),
        )


# Automatically call the method above when importing this module
load_cluster_settings()



# -*- coding: utf-8 -*-

from scipy.constants import Avogadro

from simmate.toolkit import Structure as ToolkitStructure
from simmate.database.base_data_types import (
    DatabaseTable,
    table_column,
    Spacegroup,
)


class Structure(DatabaseTable):
    class Meta:
        abstract = True

    base_info = ["structure_string"]

    structure_string = table_column.TextField(blank=True, null=True)
    """
    The core structure information, which is written to a string and in a 
    compressed format using the `from_toolkit` method. To get back to our toolkit
    structure object, use the `to_toolkit` method.
    """

    nsites = table_column.IntegerField(blank=True, null=True)
    """
    The total number of sites in the unitcell. (e.g. Y2CF2 has 5 sites)
    """

    nelements = table_column.IntegerField(blank=True, null=True)
    """
    The total number of unique elements. (e.g. Y2CF2 has 3 elements)
    """

    elements = table_column.JSONField(blank=True, null=True)
    """
    List of elements in the structure (ex: ["Y", "C", "F"])
    """

    chemical_system = table_column.CharField(max_length=25, blank=True, null=True)
    """
    the base chemical system (ex: "Y-C-F")
    
    Note: be careful when searching for elements! Running chemical_system__contains="C"
    on this field won't do what you expect -- because it will return structures
    containing Ca, Cs, Ce, Cl, and so on. If you want to search for structures
    that contain a specific element, use elements__contains='"C"' instead. The
    odd use of quotes '"C"' is required here!
    """

    density = table_column.FloatField(blank=True, null=True)
    """
    The density of the crystal in g/cm^3
    """

    density_atomic = table_column.FloatField(blank=True, null=True)
    """
    The density of atoms in the crystal in atoms/Angstom^3
    """

    volume = table_column.FloatField(blank=True, null=True)
    """
    The volume of the unitcell in Angstom^3
    
    Note: in most cases, `volume_molar` should be used instead! This is because
    volumne is highly dependent on the symmetry and the arbitray unitcell. If 
    you are truly after small volumes of the unitcell, it is likely you really 
    just want to search by spacegroup.
    """

    volume_molar = table_column.FloatField(blank=True, null=True)
    """
    The molar volume of the crystal in cm^3/mol
    """

    # The composition of the structure formatted in various ways
    # BUG: The max length here is overkill because there are many structures
    # with 8+ elements and disordered formula (e.g. "Ca2.103 N0.98")

    formula_full = table_column.CharField(
        max_length=50,
        blank=True,
        null=True,
    )
    """
    The chemical formula with elements sorted by electronegativity (ex: Li4 Fe4 P4 O16)
    """

    formula_reduced = table_column.CharField(
        max_length=50,
        blank=True,
        null=True,
    )
    """
    The reduced chemical formula. (ex: Li4Fe4P4O16 --> LiFePO4)
    """

    formula_anonymous = table_column.CharField(
        max_length=50,
        blank=True,
        null=True,
    )
    """
    An anonymized formula. Unique species are arranged in ordering of 
    amounts and assigned ascending alphabets. Useful for prototyping formulas. 
    For example, all stoichiometric perovskites have anonymized_formula ABC3.
    """

    spacegroup = table_column.ForeignKey(
        Spacegroup,
        on_delete=table_column.PROTECT,
        blank=True,
        null=True,
    )
    """
    Spacegroup information. Points to a separate database table that has additional
    columns:
    `simmate.database.base_data_types.symmetry.Spacegroup`
    """

    # The AFLOW prototype that this structure maps to.
    # TODO: this will be a relationship in the future
    # prototype = table_column.CharField(max_length=50, blank=True, null=True)

    # NOTE: extra fields for the Lattice and Sites are intentionally left out
    # in order to save on overall database size. Things such as...
    #   Lattice: matrix and then... a, b, c, alpha, beta, gamma, volume
    #   Sites: abc, xyz, properties, species/element, occupancy
    # shouldn't be queried directly. If you'd like to sort structures by these
    # criteria, you can still do this in python and pandas! Just not at the
    # SQL level

    @classmethod
    def _from_toolkit(
        cls,
        structure: ToolkitStructure,
        as_dict: bool = False,
        **kwargs,
    ):

        # BUG: This is an old line and I can't remember why I have it. Once I
        # have implemented more unittests, consider deleting. This method is
        # ment to convert to a ToolkitStructure, but the structure should
        # already be in this format...
        structure = ToolkitStructure.from_dynamic(structure)

        # OPTIMIZE: I currently store files as poscar strings for ordered structures
        # and as CIFs for disordered structures. Both of this include excess information
        # that slightly inflates file size, so I will be making a new string format in
        # the future. This will largely be based off the POSCAR format, but will
        # account for disordered structures and all limit repeated data (such as the
        # header line, "direct", listing each element/composition, etc.).
        storage_format = "POSCAR" if structure.is_ordered else "CIF"

        # OPTIMIZE
        # This attempts to match the structure to an AFLOW prototype and it is
        # by far the slowest step of loading structures to the database. Try
        # to optimize this in the future.
        # from pymatgen.analysis.prototypes import AflowPrototypeMatcher
        # prototype = AflowPrototypeMatcher().get_prototypes(structure)
        # prototype_name = prototype[0]["tags"]["mineral"] if prototype else None

        # Given a pymatgen structure object, this will return a database structure
        # object, but will NOT save it to the database yet. The kwargs input
        # is only if you inherit from this class and add extra fields.
        structure_dict = dict(
            structure_string=structure.to(fmt=storage_format),
            nsites=structure.num_sites,
            nelements=len(structure.composition),
            elements=[str(e) for e in structure.composition.elements],
            chemical_system=structure.composition.chemical_system,
            density=structure.density,
            density_atomic=structure.num_sites / structure.volume,
            volume=structure.volume,
            # 1e-27 is to convert from cubic angstroms to Liter and then 1e3 to
            # mL. Therefore this value is in mL/mol
            # OPTIMIZE: move this to a class method
            volume_molar=(structure.volume / structure.num_sites)
            * Avogadro
            * 1e-27
            * 1e3,
            spacegroup_id=structure.get_space_group_info()[1],  # OPTIMIZE
            formula_full=structure.composition.formula,
            formula_reduced=structure.composition.reduced_formula,
            formula_anonymous=structure.composition.anonymized_formula,
            # prototype=prototype_name,
            **kwargs,  # this allows subclasses to add fields with ease
        )
        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return structure_dict if as_dict else cls(**structure_dict)

    def to_toolkit(self) -> ToolkitStructure:
        """
        Converts the database object to toolkit Structure object.
        """
        return ToolkitStructure.from_database_object(self)



# -*- coding: utf-8 -*-

import os

from simmate.workflow_engine import ErrorHandler
from simmate.calculators.vasp.inputs import Incar


class InsufficientBands(ErrorHandler):
    """
    If the calculation has too few bands, we increase the number of bands by
    10% and try again.
    """

    is_monitor = True
    filename_to_check = "vasp.out"
    possible_error_messages = ["TOO FEW BANDS"]

    def correct(self, directory: str) -> str:

        # load the INCAR file to view the current settings
        incar_filename = os.path.join(directory, "INCAR")
        incar = Incar.from_file(incar_filename)

        # Grab the current number of bands. First check the INCAR and if
        # it isn't there, jump to the OUTCAR.
        if "NBANDS" in incar:
            nbands_current = incar["NBANDS"]
        else:
            outcar_filename = os.path.join(directory, "OUTCAR")
            with open(outcar_filename) as file:
                lines = file.readlines()
            # Go through the lines until we find the NBANDS. The value
            # should be at the very end of the line.
            for line in lines:
                if "NBANDS" in line:
                    nbands_current = int(line.split()[-1])
                    break
        # increase the number of bands by 10%
        nbands_new = int(nbands_current * 1.1)
        incar["NBANDS"] = nbands_new
        correction = f"switch NBANDS from {nbands_current} to {nbands_new} (+10%)"

        # rewrite the INCAR with new settings
        incar.to_file(incar_filename)

        return correction



# -*- coding: utf-8 -*-

"""
This defines the command for starting new custom project.
"""

import click


@click.command()
@click.argument(
    "project_name",
    type=click.Path(exists=False),
)
def start_project(project_name):
    """
    This creates a new folder and fills it with an example project to
    get you started with custom simmate code.
    """

    import os
    from django.core.management import call_command
    from simmate import configuration  # needed just for the filepath

    # This directory is where our template folder is located. We find this
    # by looking at the import path to see where python installed it.
    config_directory = os.path.dirname(os.path.abspath(configuration.__file__))

    # We add on "project_template" to this file path as that's our full template
    template_directory = os.path.join(config_directory, "example_project")

    # we now make the project folder using our template directory.
    # Note, we are using Django's "startproject" command even though we are just
    # copying files over. This might be overkill but it gets the job done.
    call_command("startproject", project_name, template=template_directory)

    # grab the full path to the new project for the user to see
    new_project_directory = os.path.join(os.getcwd(), project_name)

    # also navigate to the user's ~/simmate/applications.yaml file and we
    # want to add a new line at the end of it (or create the file if it isn't
    # there yet)
    from pathlib import Path

    apps_yaml = os.path.join(Path.home(), "simmate", "applications.yaml")
    new_line = "\nexample_app.apps.ExampleAppConfig"  # \n ensures a new line

    # If the file exists, we append this line to the end of the file. Otherwise,
    # we create a new file and add the line!
    with open(apps_yaml, "a+") as file:
        file.write(new_line)
    # Let the user know what we did and how to continue.
    click.echo(
        f"\n\tSuccessfully made a new project! You'll find it at {new_project_directory}\n\n"
        "\tBe sure to go through the README file in your new project.\n"
    )


# explicitly list functions so that pdoc doesn't skip them
__all__ = ["start_project"]



# -*- coding: utf-8 -*-

from typing import Callable

from simmate.calculators.vasp.inputs.incar_modifiers import (
    keyword_modifier_density,
    keyword_modifier_density_a,
    keyword_modifier_density_b,
    keyword_modifier_density_c,
    keyword_modifier_per_atom,
    keyword_modifier_smart_magmom,
    keyword_modifier_smart_lmaxmix,
    keyword_modifier_smart_ldau,
    keyword_modifier_smart_ismear,
)


class Incar(dict):
    """
    INCAR object for reading and writing INCAR files. This behaves exactly like
    a python dictionary, but has a few extra checks and methods attached to it.
    You can pass it a dictionary or initialize it just like you would dict(kwargs).
    You can consider the dict(kwargs) as equivalent to Incar(parameters).

    If you want a given setting to be dependent on the structure or dynamically
    determined, then we implement these modifiers. This would enable us to
    do things like ENCUT__per_atom or NGZF__density. We can even have more complex
    modifiers like multiple_keywords__smart_ldau which signals that our
    "smart_ldau" modifier introduces more than one new setting to the INCAR, such
    as LDAUJ, LDAUU, LDAUL, LDAUTYPE, and LDAUPRINT.
    """

    def __init__(self, **kwargs):

        # The kwargs are a dictionary of parameters (e.g. {"PREC": "accurate"})

        # Establish the dictionary with the given input
        # This is the same as doing dict(**kwargs)
        super().__init__(**kwargs)

        # Based on the kwargs, go through and make sure each parameter is in
        # it's expected datatype. Thus users can supply strings in the typical
        # VASP format that will be convert to python types here.
        # OPTIMIZE -- would it be faster if I only did this on from_file init?
        # and therefore assumed the user to initialize this with proper python
        # datatypes and formatting? Also will this behave properly if the value
        # is already in the correct format?
        for parameter, value in self.items():

            # parameters may have tags like "__density" added onto them. We don't
            # convert to the datatype yet, but instead wait until a structure
            # is provided (when the to_file or str methods are called below).
            # These are defined in python and should already be in the correct
            # data format.
            if "__" in parameter:
                self.update({parameter: value})

            # Otherwise, we need to convert values to the proper python datatype
            # because we might be reading from a file where everything is a string
            else:
                formatted_value = self._str_to_datatype(parameter, value)
                self.update({parameter: formatted_value})

        # SPECIAL CASE
        # If you have LSORBIT=True or LNONCOLLINEAR=True, the MAGMOM class needs
        # to take on a different form. Instead of a single MAGMOM being set as
        # a float, it's now a vector (x,y,z). For example:
        #   MAGMOM = x1 y1 z1    x2 y2 z2    x3 y3 z3 ...
        #       [...instead of...]
        #   MAGMOM = value1 value2 value3 ...
        # Check if MAGMOM is set along with LSORBIT or LNONCOLLINEAR being True
        if ("MAGMOM" in kwargs) and (
            kwargs.get("LSORBIT") or kwargs.get("LNONCOLLINEAR")
        ):
            # convert the MAGMOM value from...
            #   [x1,y1,z1,x2,y2,z2,x3,y3,z3] (this can be a string too)
            #   to...
            #   [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3]]
            old_format = kwargs["MAGMOM"]
            new_format = [old_format[i : i + 3] for i in range(0, len(old_format), 3)]
            # Or If I wanted to do the same thing with numpy:
            #   new_format = numpy.array_split(old_format, len(old_format) // 3)
            # now update the dictionary with this value
            self.update({"MAGMOM": new_format})

    def to_evaluated_str(self, structure=None):

        # Let's start with an empty string and build from there
        final_str = ""

        # First we need to iterate through all parameters and check if we have
        # ones that are structure-specific. For example, we would need to
        # evaluate "ENCUT__per_atom". We go through all these and collect the
        # parameters into a final settings list.
        final_settings = {}
        for parameter, value in self.items():

            # if there is no modifier attached to the parameter, we just keep it as-is
            if "__" not in parameter:
                final_settings[parameter] = value

            # Otherwise we have a modifier like "__density" and need to evaluate it
            else:

                # make sure we have a structure supplied because all modifiers
                # require one.
                if not structure:
                    raise Exception(
                        "It looks like you used a keyword modifier but didn't "
                        f"supply a structure! If you want to use {parameter}, "
                        "then you need to make sure you provide a structure so "
                        "that the modifier can be evaluated."
                    )

                # separate the input into the base parameter and modifier.
                # This also overwrites what our paramter value is.
                parameter, modifier_tag = parameter.split("__")

                # check that this class has this modifier supported. It should
                # be a method named "keyword_modifier_mymodifier".
                # If everything looks good, we grab the modifier function
                modifier_fxn_name = "keyword_modifier_" + modifier_tag
                if hasattr(self, modifier_fxn_name):
                    modifier_fxn = getattr(self, modifier_fxn_name)
                else:
                    raise AttributeError(
                        """
                        It looks like you used a keyword modifier that hasn't
                        been defined yet! If you want something like ENCUT__smart_encut,
                        then you need to make sure there is a keyword_modifier_smart_encut
                        method available.
                        """
                    )

                # now that we have the modifier function, let's use it to update
                # our value for this keyword.
                value = modifier_fxn(structure, value)

                # sometimes the modifier returns None. In this case we don't
                # set anything in the INCAR, but leave it to the programs
                # default value.
                if not value:
                    continue
                # BUG: Are there cases where None is return but we still want
                # to write it to the INCAR? If so, it'd be skipped here.

                # if the "parameter" is actually "multiple_keywords", then we
                # have our actual parameters as a dictionary. We need to
                # pull these out of the "value" we have.
                if parameter == "multiple_keywords":
                    for subparameter, subvalue in value.items():
                        final_settings[subparameter] = subvalue

                # otherwise we were just given back an update value
                else:
                    final_settings[parameter] = value

        # Now that we have all of our parameters evaluated for the structure, we
        # iterate through each parameter and its set value. Each one will be
        # put on a separate line.
        for parameter, value in final_settings.items():

            # let's start by adding the parameter key to our output
            # It will be followed by an equal sign to separate it's value
            final_str += parameter + " = "

            # If we have a value that is a list (or list of lists)
            if isinstance(value, list):

                # check if we have a list of lists
                # Take MAGMOM with LSORBIT=True as an example, where we convert...
                #   [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3]]
                # to...
                # "x1 y1 z1 x2 y2 z2 x3 y3 z3"
                if isinstance(value[0], list):
                    final_str += " ".join(str(n) for xyz in value for n in xyz)

                # otherwise we just have a list of items
                # converts [1, 2, 3] to "1 2 3"
                else:
                    final_str += " ".join(str(n) for n in value)

            # if it's not a list, we can just save the string of the value
            else:
                final_str += str(value)

            # we want to start each parameter on a new line, so we need to
            # add this at the end of each. There will be an extra new line at
            # the end of the file, but that's alright
            final_str += "\n"

        # we now have our final string and can return it!
        return final_str

    def to_file(self, filename="INCAR", structure=None):
        """
        Write Incar to a file.
        Args:
            filename (str): filename to write to.
        """
        # we just take the string format and put it in a file
        with open(filename, "w") as file:
            file.write(self.to_evaluated_str(structure=structure))

    @staticmethod
    def from_file(filename="INCAR"):
        """
        Reads an Incar object from a file.
        Args:
            filename (str): Filename for file
        Returns:
            Incar object
        """
        # open the file, grab the lines, and then close it
        with open(filename) as file:
            lines = file.readlines()

        # store parameters in this dictionary
        parameters = {}
        # go through line by line
        for line in lines:
            # If the line starts with a # then its a comment and we should skip.
            # It also could be an empty line that we should skip,
            if line.startswith("#") or line.startswith("\n"):
                continue
            # multiple parameters on a single line are separated by a semicolon
            for sub_line in line.split(";"):
                # the PARAMETER and VALUE are separated by equal sign
                parameter, value = sub_line.split("=")
                # we can use the parameter/value to update our dictionary. The
                # last thing we do is remove leading/trailing whitespace with strip()
                parameters[parameter.strip()] = value.strip()

        # return the final dictionary as an Incar object
        return Incar(**parameters)

    @staticmethod
    def _str_to_datatype(parameter, value):
        """
        When given a vasp parameter and it's value as a string, this helper
        function will use the key (parameter) to determine how to convert the
        val string to the proper python datatype (int, float, bool, list...).
        I have the most common keys mapped out, but if a parameter is given that
        isn't mapped, I simply leave it as a string.
        """

        # I outline the most common keys to what their expected data types are.

        # OPTIMIZE -- I should set these elsewhere so that these lists are not
        # initialized every time I call this function. Maybe have a dictionary
        # of {Parameter: Value_datatype} in the main enviornment for use.

        vector_list_keys = (
            # "MAGMOM",  # depends on other args -- see notes in init
            "DIPOL",
        )

        float_list_keys = (
            "LDAUU",
            "LDAUJ",
            "MAGMOM",  # depends on other args -- see notes in init
            "LANGEVIN_GAMMA",
            "QUAD_EFG",
            "EINT",
        )
        int_list_keys = (
            "LDAUL",
            "LDAUJ",
            "EINT",
        )

        bool_keys = (
            "LDAU",
            "LWAVE",
            "LSCALU",
            "LCHARG",
            "LPLANE",
            "LUSE_VDW",
            "LHFCALC",
            "ADDGRID",
            "LSORBIT",
            "LNONCOLLINEAR",
            "KGAMMA",
        )
        float_keys = (
            "EDIFF",
            "SIGMA",
            "TIME",
            "ENCUTFOCK",
            "HFSCREEN",
            "POTIM",
            "EDIFFG",
            "AGGAC",
            "PARAM1",
            "PARAM2",
            "KSPACING",
            "SYMPREC",
            "AMIX",
            "BMIX",
            "AMIN",
            "SMASS",
            "AMIX_MAG",
            "BMIX_MAG",
        )
        int_keys = (
            "NSW",
            "NBANDS",
            "NELMIN",
            "ISIF",
            "IBRION",
            "ISPIN",
            "ICHARG",
            "NELM",
            "ISMEAR",
            "NPAR",
            "LDAUPRINT",
            "LMAXMIX",
            "ENCUT",
            "NSIM",
            "NKRED",
            "NUPDOWN",
            "ISPIND",
            "LDAUTYPE",
            "IVDW",
            "ISTART",
            "NELMDL",
            "IMIX",
            "ISYM",
        )

        # If the value is not a string, then assume we are already in the
        # correct format. Note, an incorrect format will throw an error
        # somewhere below, which may be tricky for beginners to traceback.
        if not isinstance(value, str):
            return value

        # if the parameter is in int_keys
        if parameter in int_keys:
            # sometimes "1." was written to indicate an integer so check for
            # this and remove it if needed.
            if value[-1] == ".":
                value = value[:-1]
            # return the value integer
            return int(value)

        # if the parameter is in float_keys, we convert value to a float
        elif parameter in float_keys:
            # return the value float
            return float(value)

        # if the parameter is in bool_keys
        elif parameter in bool_keys:
            # Python is weird where bool("FALSE") will return True... So I need
            # to convert the string to lowercase and read it to know what to
            # return here.
            if "t" in value.lower():
                return True
            elif "f" in value.lower():
                return False

        # if the parameter is in vector_list_keys
        # These vectors are always floats
        elif parameter in vector_list_keys:
            # convert a string of...
            #   "x1 y1 z1 x2 y2 z2 x3 y3 z3"
            # to...
            #   [x1,y1,z1,x2,y2,z2,x3,y3,z3] (list of floats)
            # and then to...
            #   [[x1,y1,z1],[x2,y2,z2],[x3,y3,z3]]
            value = [float(item) for item in value.split()]
            return [value[i : i + 3] for i in range(0, len(value), 3)]

        # if the parameter is in float_list_keys
        elif parameter in float_list_keys:
            final_list = []
            for item in value.split():
                # Sometimes, the values are given as "3*0.1 2*0.5" where the "*"
                # means to include that value that many times. For example, this
                # input would be the same as "0.1 0.1 0.1 0.5 0.5". We need to
                # account for this when parsing.
                if "*" in item:
                    nsubitems, subitem = item.split("*")
                    for n in range(int(nsubitems)):
                        final_list.append(float(subitem))
                else:
                    final_list.append(float(item))
            return final_list

        # if the parameter is in int_list_keys
        elif parameter in int_list_keys:
            return [int(item) for item in value.split()]

        # If it is not in the common keys listed, just leave it as a string.
        else:
            return value

    def compare_incars(self, other_incar):
        """
        Compares two Incars and indicates which parameters are the same and
        which are not. Useful for checking whether two runs were done using
        the same parameters.
        Args:
            other (Incar): The other Incar object to compare to.
        Returns:
            Dict of the following format:
            {"Same" : parameters_that_are_the_same,
            "Different": parameters_that_are_different}
            The value of the other_incar is returned for the paramters that
            are different are returned as {Parameter: (incar_value, other_incar_value)}
            where None is used as a placeholder. For the same parameters, the
            dictionary is returned as {Parameter: shared_value}
        """

        same_parameters = {}
        different_parameters = {}

        # make a collection of all unique keys used by the two dictionaries
        # I convert to list() so I can use set() below.
        # OPTIMIZE -- is there a better way to do this?
        parameters1 = list(self.keys())
        parameters2 = list(other_incar.keys())

        # iterate through each unique parameter
        for parameter in set(parameters1 + parameters2):

            # try to grab the value from both Incar objects
            # If it doesn't have the value, None will be provided
            value1 = self.get(parameter)
            value2 = other_incar.get(parameter)

            # compare the two and store
            if value1 == value2:
                # it doesn't matter which value I grab so I just use value1
                same_parameters[parameter] = value1
            # if they are different...
            else:
                different_parameters[parameter] = (value1, value2)

        return {"Same": same_parameters, "Different": different_parameters}

    def __add__(self, other):
        """
        Add all the values of another INCAR object to this object.
        Facilitates the use of "standard" INCARs.
        """
        # start by copying this incar's settings
        new_parameters = dict(self.items())

        # now iterate through the other incar parameters to incorporate them
        for parameter, value in other.items():

            # If both have a given parameter, make sure they are the same values
            if parameter in self and value != self[parameter]:
                # If not, we shouldn't allow combining these Incars as it can
                # lead to undesired results.
                raise ValueError(
                    f"Incars have conflicting values! One conflict is with {parameter} "
                    f"where one incar has {self[parameter]} and the other has {value}"
                )
            # otherwise just set the value
            else:
                new_parameters[parameter] = value

        # now return the new Incar
        return Incar(**new_parameters)

    @classmethod
    def add_keyword_modifier(cls, keyword_modifier: Callable):
        """
        Dynamically sets a keyword_modifier function to the Incar class. When
        loaded onto the Incar class, you can then use the modifiers name to
        set a parameter based on the input structure.

        Names of modifiers MUST start with "keyword_modifier_" and
        have unique endings (i.e. you can't have 2 "per_atom" modifiers).

        For example, in order to allow settings like `ENCUT__per_atom`, we use
        a `keyword_modifier_per_atom` function and add it to this class:

        ``` python

        # create your modifier logic as a function
        def keyword_modifier_double_per_atom(structure, per_atom_value):
            return per_atom_value * structure.num_sites * 2

        # add it to the Incar class to make it active
        Incar.add_keyword_modifier(keyword_modifier_double_per_atom)

        # You can not use the modifier in incar logic!
        incar = Incar(ENCUT__double_per_atom=1e-5)
        ```
        """

        # first ensure that the modifier does not conflict with an existing one
        if hasattr(cls, keyword_modifier.__name__):
            raise Exception(
                "The Incar class already has a modifier with this name. "
                "Please use a different modifier name to avoid conflicts."
            )

        # add the function/callable as a static method to this class.
        # This line of code is based on...
        #   https://stackoverflow.com/questions/17929543/
        setattr(cls, keyword_modifier.__name__, staticmethod(keyword_modifier))


# set some default keyword modifiers
for modifier in [
    keyword_modifier_density,
    keyword_modifier_density_a,
    keyword_modifier_density_b,
    keyword_modifier_density_c,
    keyword_modifier_per_atom,
    keyword_modifier_smart_magmom,
    keyword_modifier_smart_lmaxmix,
    keyword_modifier_smart_ldau,
    keyword_modifier_smart_ismear,
]:
    Incar.add_keyword_modifier(modifier)



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.tasks.static_energy.materials_project import (
    MatprojStaticEnergy,
)


class MatprojELF(MatprojStaticEnergy):
    """
    Runs a static energy calculation under settings from the Materials Project
    and also writes the electron localization function (to ELFCAR).
    """

    incar = MatprojStaticEnergy.incar.copy()
    incar.update(
        LELF=True,  # writes ELFCAR
        NPAR=1,  # must be set if LELF is set to True
        # BUG: if NPAR conflicts with INCAR_parallel_settings config this
        # fails and tells the user to specify a setting
    )


# -----------------------------------------------------------------------------

# Below are old notes from a script I used to workup ELF. These might be useful
# if I want to update the workup method of this task. For now, no extra workup
# is done for this calculation as typically the ELFCAR is just used for
# visualization.

# -----------------------------------------------------------------------------

# # import the VASP results
# from pymatgen.io.vasp.outputs import Vasprun

# xmlReader = Vasprun(
#     filename="vasprun.xml",
#     parse_dos=True,
#     parse_eigen=True,
#     parse_projected_eigen=False,
#     parse_potcar_file=True,
#     exception_on_bad_xml=True,
# )


# # grab the info we want
# final_structure = xmlReader.structures[-1]  # or Structure.from_file('CONTCAR')
# final_energy = (
#     xmlReader.final_energy / final_structure.num_sites
# )  #!!! convert this to per_atom!
# converged = xmlReader.converged

# from pymatgen.io.vasp.outputs import Elfcar

# elfcar = Elfcar.from_file("ELFCAR")

# alphacar = elfcar.get_alpha()
# alphacar.write_file("ALPHACAR.vasp")

# from pymatgen.io.vasp.outputs import Chgcar

# chgcar = Chgcar.from_file("CHGCAR")

# # scale data of the Elfcar to the size of Chgcar using Kronecker product
# import numpy as np

# scale = [
#     int(c / e) for c, e in zip(chgcar.data["total"].shape, elfcar.data["total"].shape)
# ]
# elfcar_scaled = np.kron(elfcar.data["total"], np.ones(scale))
# from scipy.ndimage import gaussian_filter  #!!! This is for smoothing and optional

# elfcar_scaled = gaussian_filter(elfcar_scaled, sigma=0.75)
# # from scipy.linalg import kron
# # elfcar_scaled = kron(elfcar.data["total"], chgcar.data["total"])
# elfcar_scaled = Elfcar(elfcar.structure, {"total": elfcar_scaled})
# elfcar_scaled.write_file("ELFCAR_scaled.vasp")

# alphacar_scaled = elfcar_scaled.get_alpha()
# alphacar_scaled.write_file("ALPHACAR_scaled.vasp")

# # also write the crystal structure because it may differ from the input
# elfcar.structure.to(filename="primitive_structure.cif")

# # For 3d integrations:
# # https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.tplquad.html
# # def f(x,y,z):
# #     return chgcar.data['total'][0][1][2]
# # tplquad(f, 1, 2, lambda x: 2, lambda x: 3, lambda x, y: 0, lambda x, y: 1)
# # OR
# # https://stackoverflow.com/questions/47415122/calculate-the-volume-of-3d-plot


# import numpy as np
# import itertools


# def get_integrated_total(elfcar, ind, radius, nbins=1):
#     """
#     Get integrated difference of atom index ind up to radius. This can be
#     an extremely computationally intensive process, depending on how many
#     grid points are in the VolumetricData.
#     Args:
#         ind (int): Index of atom.
#         radius (float): Radius of integration.
#         nbins (int): Number of bins. Defaults to 1. This allows one to
#             obtain the charge integration up to a list of the cumulative
#             charge integration values for radii for [radius/nbins,
#             2 * radius/nbins, ....].
#     Returns:
#         Differential integrated charge as a np array of [[radius, value],
#         ...]. Format is for ease of plotting. E.g., plt.plot(data[:,0],
#         data[:,1])
#     """
#     struct = elfcar.structure
#     a = elfcar.dim
#     if (
#         ind not in elfcar._distance_matrix
#         or elfcar._distance_matrix[ind]["max_radius"] < radius
#     ):
#         coords = []
#         for (x, y, z) in itertools.product(*[list(range(i)) for i in a]):
#             coords.append([x / a[0], y / a[1], z / a[2]])
#         sites_dist = struct.lattice.get_points_in_sphere(
#             coords, struct[ind].coords, radius
#         )
#         elfcar._distance_matrix[ind] = {
#             "max_radius": radius,
#             "data": np.array(sites_dist),
#         }

#     data = elfcar._distance_matrix[ind]["data"]

#     # Use boolean indexing to find all charges within the desired distance.
#     inds = data[:, 1] <= radius
#     dists = data[inds, 1]
#     data_inds = np.rint(
#         np.mod(list(data[inds, 0]), 1) * np.tile(a, (len(dists), 1))
#     ).astype(int)
#     vals = [elfcar.data["total"][x, y, z] for x, y, z in data_inds]  #!!! diff to total

#     hist, edges = np.histogram(dists, bins=nbins, range=[0, radius], weights=vals)
#     data = np.zeros((nbins, 2))
#     data[:, 0] = edges[1:]
#     data[:, 1] = [sum(hist[0 : i + 1]) / elfcar.ngridpts for i in range(nbins)]
#     return data


# data = get_integrated_total(chgcar, ind=0, radius=2.5, nbins=100)

# import matplotlib.pyplot as plt

# plt.plot(data[:, 0], data[:, 1])



# -*- coding: utf-8 -*-

"""
Structure Relaxations are made up of a series of ionic steps -- where each
ionic step can be thought of as a single energy calculation. This yields
energy, forces, and lattice stress for a given structure. In many cases,
users won't need all of this data because they're only after the final
structure. However, all ionic steps are useful in the case of confirming 
convergence as well as machine learning applications. We therefore store these
here.

This module helps you create tables to results from structure relaxations
(aka geometry optimizations). This will store all ionic steps and the forces/stress
associated with each step.

When creating new tables for Relaxations, you should use the `Relaxation.create_subclasses` method, which helps remove all the 
boilerplate code needed. For Django users, it may be tricky to understand what's
happening behind the scenes, so here's an example:
    
These two lines...

``` python 
from simmate.database.base_data_types import Relaxation

ExampleRelaxation, ExampleIonicStep = Relaxation.create_subclasses(
    "Example",
    module=__name__,
)
```

... do exactly the same thing as all of these lines...

``` python
from simmate.database.base_data_types import table_column
from simmate.database.base_data_types import IonicStep, Relaxation

class ExampleIonicStep(IonicStep):
    relaxation = table_column.ForeignKey(
        "ExampleRelaxation",  # in quotes becuase this is defined below
        on_delete=table_column.CASCADE,
        related_name="structures",
    )

class ExampleRelaxation(Relaxation):
    structure_start = table_column.OneToOneField(
        ExampleIonicStep,
        on_delete=table_column.CASCADE,
        related_name="relaxations_as_start",
        blank=True,
        null=True,
    )
    structure_final = table_column.OneToOneField(
        ExampleIonicStep,
        on_delete=table_column.CASCADE,
        related_name="relaxations_as_final",
        blank=True,
        null=True,
    )
```

Note there are two tables involved. One stores all of the ionic steps, and the
other connects all ionic steps to a specific calculation and result.

"""

import os

from simmate.database.base_data_types import (
    table_column,
    Structure,
    Forces,
    Thermodynamics,
    Calculation,
)

from plotly.subplots import make_subplots
import plotly.graph_objects as plotly_go

from typing import List
from pymatgen.io.vasp.outputs import Vasprun


class Relaxation(Structure, Thermodynamics, Calculation):
    """
    This table holds all data from a structure relaxation and also links to
    IonicStep table which holds all of the structure/energy/forces for each
    ionic step.

    WARNING: The Structure stored in this table here is the source structure
    until the calculation completes. After completed, the structure will
    be updated to the final structure. If you wish to ensure you're accessing
    the correct structure, use the `structure_final` attribute, which gives
    the final IonicStep.
    """

    class Meta:
        abstract = True
        app_label = "workflows"

    base_info = (
        [
            "band_gap",
            "is_gap_direct",
            "energy_fermi",
            "conduction_band_minimum",
            "valence_band_maximum",
        ]
        + Structure.base_info
        + Calculation.base_info
    )

    # OPTIMIZE: should I include this electronic data?
    # This data here is something we only get for the final structure, so it
    # may make sense to move this data into the IonicStepStructure table (and
    # allow null values for non-final steps). I instead keep this here because
    # I don't want columns above that are largely empty.
    # Note: all entries are optional because there is no guaruntee the calculation
    # finishes successfully

    band_gap = table_column.FloatField(blank=True, null=True)
    """
    The band gap energy in eV.
    """

    is_gap_direct = table_column.BooleanField(blank=True, null=True)
    """
    Whether the band gap is direct or indirect.
    """

    energy_fermi = table_column.FloatField(blank=True, null=True)
    """
    The Fermi energy in eV.
    """

    conduction_band_minimum = table_column.FloatField(blank=True, null=True)
    """
    The conduction band minimum in eV.
    """

    valence_band_maximum = table_column.FloatField(blank=True, null=True)
    """
    The valence band maximum in eV.
    """

    volume_change = table_column.FloatField(blank=True, null=True)
    """
    The percent volume change during the relaxation. This is useful for checking
    if Pulay Stress may be significant or if the structure was highly unreasonable.
    We store this as a ratio relative to the starting structure:
    ```
    (volume_final - volume_start) / volume_start
    ```
    """

    """ Relationships """
    # structure_start --> points to first (0) IonicStep
    # structure_final --> points to final IonicStep
    # structures --> gives list of all IonicSteps

    @classmethod
    def create_subclasses(cls, name: str, module: str, **extra_columns):
        """
        Dynamically creates a subclass of Relaxation as well as a separate IonicStep
        table for it. These tables are linked together.

        Example use:

        ``` python
        from simmate.database.base_data_types import Relaxation

        ExampleRelaxation, ExampleIonicStep = Relaxation.create_subclasses(
            "Example",
            module=__name__,
        )
        ```

        #### Parameters

        - `name` :
            The prefix name of the subclasses that are output. "Relaxation" and
            "IonicStep" will be attached to the end of this prefix.
        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.
        - `**extra_columns` :
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        #### Returns

        - `NewRelaxationClass` :
            A subclass of Relaxation.
        - `NewIonicStepClass`:
            A subclass of IonicStep.

        """

        # For convience, we add columns that point to the start and end structures
        NewRelaxationClass = cls.create_subclass(
            f"{name}Relaxation",
            structure_start=table_column.OneToOneField(
                f"{name}IonicStep",
                on_delete=table_column.CASCADE,
                related_name="relaxations_as_start",
                blank=True,
                null=True,
            ),
            structure_final=table_column.OneToOneField(
                f"{name}IonicStep",
                on_delete=table_column.CASCADE,
                related_name="relaxations_as_final",
                blank=True,
                null=True,
            ),
            module=module,
            **extra_columns,
        )

        NewIonicStepClass = IonicStep.create_subclass_from_relaxation(
            name,
            NewRelaxationClass,
            module=module,
            **extra_columns,
        )

        # we now have a new child class and avoided writing some boilerplate code!
        return NewRelaxationClass, NewIonicStepClass

    @classmethod
    def from_directory(cls, directory: str):
        # I assume the directory is from a vasp calculation, but I need to update
        # this when I begin adding new calculators.
        vasprun_filename = os.path.join(directory, "vasprun.xml")
        vasprun = Vasprun(vasprun_filename)
        return cls.from_vasp_run(vasprun)

    @classmethod
    def from_vasp_run(cls, vasprun: Vasprun):

        # Make the relaxation entry. Note we need to save this to the database
        # in order to link/save the ionic steps below. We save the structure
        # as the final one in the calculation.
        # Note, the information does not matter at this point because it will be
        # populated below
        relaxation = cls.from_toolkit(structure=vasprun.structures[-1])
        # TODO: need to pull prefect_flow_run_id from metadata file.

        # Now we have the relaxation data all loaded and can save it to the database
        relaxation.save()

        # Save the rest of the results using the update method from this class
        relaxation.update_from_vasp_run(
            vasprun=vasprun,
            corrections=[],
            directory=None,
        )
        # TODO: load corrections/directory from the metadata and corrections files.

        return relaxation

    def update_from_vasp_run(
        self,
        vasprun: Vasprun,
        corrections: List,
        directory: str,
    ):
        """
        Given a Vasprun object from a finished relaxation, this will update the
        Relaxation table entry and the corresponding IonicStep entries.

        #### Parameters

        vasprun :
            The final Vasprun object from the relaxation outputs.
        corrections :
            List of errors and corrections applied to during the relaxation.
        directory :
            name of the directory that relaxation was ran in. This is only used
            to reference the archive file if it's ever needed again.
        """

        # The data is actually easier to access as a dictionary and everything
        # we need is stored under the "output" key
        data = vasprun.as_dict()["output"]

        # The only other data we need to grab is the list of structures. We can
        # pull the structure for each ionic step from the vasprun class directly.
        structures = vasprun.structures

        # Now let's iterate through the ionic steps and save these to the database.
        for number, (structure, ionic_step) in enumerate(
            zip(structures, data["ionic_steps"])
        ):
            # first pull all the data together and save it to the database. We
            # are saving this to an IonicStepStructure datatable. To access this
            # model, we look need to use "structures.model".
            structure = self.structures.model.from_toolkit(
                number=number,
                structure=structure,
                energy=ionic_step["e_wo_entrp"],
                site_forces=ionic_step["forces"],
                lattice_stress=ionic_step["stress"],
                relaxation=self,  # this links the structure to this relaxation
            )
            structure.save()

            # If this is the first structure, we want to link it as such
            if number == 0:
                self.structure_start_id = structure.id
            # and same for the final structure. Note, we can't use elif becuase
            # there's a chance the start/end structure are the same, which occurs
            # when the starting structure is found to be relaxed already.
            if number == len(structures) - 1:
                self.structure_final_id = structure.id

        # update our relaxation entry with new data
        self.update_from_toolkit(
            # use the final ionic setup for the structure and energy
            structure=structures[-1],
            energy=data["ionic_steps"][-1]["e_wo_entrp"],
            # calculate extra data for storing
            volume_change=structures[-1].volume
            - structures[0].volume / structures[0].volume,
            # There is also extra data for the final structure that we save directly
            # in the relaxation table.  We use .get() in case the key isn't provided.
            band_gap=data.get("bandgap"),
            is_gap_direct=data.get("is_gap_direct"),
            energy_fermi=data.get("efermi"),
            conduction_band_minimum=data.get("cbm"),
            valence_band_maximum=data.get("vbm"),
            # lastly, we also want to save the corrections made and directory it ran in
            corrections=corrections,
            directory=directory,
        )

    def get_convergence_plot(self):

        # Grab the calculation's structure and convert it to a dataframe
        structures_dataframe = self.structures.order_by("number").to_dataframe()

        # We will be making a figure that consists of 3 stacked subplots that
        # all share the x-axis of ionic_step_number
        figure = make_subplots(
            rows=3,
            cols=1,
            shared_xaxes=True,
        )

        # Generate a plot for Energy (per atom)
        figure.add_trace(
            plotly_go.Scatter(
                x=structures_dataframe["number"],
                y=structures_dataframe["energy_per_atom"],
            ),
            row=1,
            col=1,
        )

        # Generate a plot for Forces (norm per atom)
        figure.add_trace(
            plotly_go.Scatter(
                x=structures_dataframe["number"],
                y=structures_dataframe["site_forces_norm_per_atom"],
            ),
            row=2,
            col=1,
        )

        # Generate a plot for Stress (norm per atom)
        figure.add_trace(
            plotly_go.Scatter(
                x=structures_dataframe["number"],
                y=structures_dataframe["lattice_stress_norm_per_atom"],
            ),
            row=3,
            col=1,
        )

        # Now let's clean up some formatting and add the axes titles
        figure.update_layout(
            width=600,
            height=600,
            showlegend=False,
            xaxis3_title="Ionic Step (#)",
            yaxis_title="Energy (eV/atom)",
            yaxis2_title="Site Forces",
            yaxis3_title="Lattice Stress",
        )

        # we return the figure object for the user
        return figure

    def view_convergence_plot(self):
        figure = self.get_convergence_plot()
        figure.show(renderer="browser")

    def get_convergence_plot_html(self):
        # Make the convergence figure and convert it to an html div
        figure_convergence = self.get_convergence_plot()
        figure_convergence_html = figure_convergence.to_html(
            full_html=False,
            include_plotlyjs=False,
        )
        return figure_convergence_html


class IonicStep(Structure, Thermodynamics, Forces):
    """
    This database table that holds the data for each ionic step of a relaxation.

    An ionic step can be viewed as a static energy calculation, but we keep these
    results separate because:

        1. Pulay stress can make these energies/forces inaccurate
        2. These results each have an associated Relaxation and ionic step number

    You will likely never access this table directly. Instead, data is better
    accessed through the `structures` attribute on appropiate Relaxation table.

    For example:

    ``` python
    from simmate.database import connect
    from simmate.database.workflow_results import MITRelaxation

    # grab your desired relaxation
    relax = MITRelaxation.objects.get(id=1)

    # grab the associated ionic steps
    ionic_steps = relax.structures.all()
    ```

    Note: we assume that only converged data is being stored! So there is no
    "converged_electronic" column here. ErrorHandlers and workups should
    ensure convergence.
    """

    class Meta:
        abstract = True
        app_label = "workflows"

    base_info = (
        ["number"] + Structure.base_info + Thermodynamics.base_info + Forces.base_info
    )

    number = table_column.IntegerField()
    """
    The ionic step number within the relaxation. This starts counting from 0.
    """

    """ Relationships """
    # each of these will map to a Relaxation, so you should typically access this
    # data through that class. The exception to this is when you want all ionic
    # steps accross many relaxations for a machine learning input.
    # These relationship can be found via...
    #   relaxations
    #   relaxations_as_start
    #   relaxations_as_final

    @classmethod
    def create_subclass_from_relaxation(
        cls,
        name: str,
        relaxation: Relaxation,
        module: str,
        **extra_columns,
    ):
        """
        Dynamically creates a subclass of IonicStep and links it to the Relaxation
        table.

        This method should NOT be called directly because it is instead used by
        `Relaxation.create_subclasses`.

        #### Parameters

        - `name` :
            Name of the subclass that is output.
        - `relaxation` :
            Relaxation table that these ionic steps should be associated with.
        - `module` :
            name of the module this subclass should be associated with. Typically,
            you should pass __name__ to this.
        - `**extra_columns` :
            Additional columns to add to the table. The keyword will be the
            column name and the value should match django options
            (e.g. table_column.FloatField())

        Returns
        -------
        - `NewClass` :
            A subclass of IonicStep.

        """

        # All structures in this table come from relaxation calculations, where
        # there can be many structures (one for each ionic steps) linked to a
        # single relaxation. This means the start structure, end structure, and
        # those structure in-between are stored together here.
        # Therefore, there's just a simple column stating which relaxation it
        # belongs to.
        NewClass = cls.create_subclass(
            f"{name}IonicStep",
            relaxation=table_column.ForeignKey(
                relaxation,
                on_delete=table_column.CASCADE,
                related_name="structures",
            ),
            module=module,
            **extra_columns,
        )

        # we now have a new child class and avoided writing some boilerplate code!
        return NewClass



# -*- coding: utf-8 -*-

from simmate.workflow_engine import Workflow
from simmate.calculators.vasp.tasks.relaxation import (
    MITRelaxation as MITRelaxationTask,
)
from simmate.calculators.vasp.database.relaxation import (
    MITRelaxation as MITRelaxationResults,
)


class Relaxation__Vasp__Mit(Workflow):
    s3task = MITRelaxationTask
    database_table = MITRelaxationResults
    description_doc_short = "uses MIT Project settings"



# -*- coding: utf-8 -*-

import os
import shutil

from django.apps import apps
from django.core.management import call_command

from simmate.configuration.django.settings import DATABASES

# Lists off which apps to update/create. By default, I do all apps that are installed
# so this list is grabbed directly from django. I also grab the CUSTOM_APPS to
# check for user-installed applications.
APPS_TO_MIGRATE = list(apps.app_configs.keys())


def update_database(apps_to_migrate=APPS_TO_MIGRATE):

    # execute the following commands to update the database
    call_command("makemigrations", *apps_to_migrate)
    call_command("migrate")


def reset_database(apps_to_migrate=APPS_TO_MIGRATE, use_prebuilt=False):

    # BUG: Why doesn't call_command("flush") do this? How is it different?

    # BUG: this is only for SQLite3
    # Consider wrapping the django-extensions function for this instead:
    #   https://django-extensions.readthedocs.io/en/latest/reset_db.html
    # An example command to call this (when django-extensions is installed) is...
    #   django-admin reset_db --settings=simmate.configuration.django.settings
    # Note: this does not remove migration files or reapply migrating after
    # Make sure the backend is using SQLite3 as this is the only allowed format
    using_sqlite = DATABASES["default"]["ENGINE"] == "django.db.backends.sqlite3"
    if not using_sqlite:
        print(
            "WARNING: reseting your database does not clear your database tables "
            "for non-sqlite3 backends. Make sure you only use this function when "
            "initially building your database and not after."
        )

    # grab the location of the database file. I assume the default database for now.
    db_filename = DATABASES["default"]["NAME"]

    # delete the sqlite3 database file if it exists
    if os.path.exists(db_filename):
        os.remove(db_filename)

    # go through each app directory and delete all folders named 'migrations'
    for app_name, app_config in apps.app_configs.items():

        # Skip if the app was not requested
        if app_config.label not in apps_to_migrate:
            continue

        migration_dir = os.path.join(app_config.path, "migrations")
        if os.path.exists(migration_dir):
            shutil.rmtree(migration_dir)
            continue

    # now update the database based on the registered models
    update_database(apps_to_migrate)

    # instead of building the database from scratch, we instead download a
    # prebuilt database file.
    if using_sqlite and use_prebuilt:
        from simmate.database.third_parties import load_default_sqlite3_build

        load_default_sqlite3_build()

    # Otherwise we leave the empty database.
    # Because this is our first time building the database, we also want to
    # load the Spacegroup metadata for us to query Structures by.
    else:
        from simmate.database.base_data_types import Spacegroup

        Spacegroup._load_database_from_toolkit()


def dump_database_to_json(filename="database_dump.json", exclude=[]):

    # execute the following commands to write the database to a json file
    call_command("dumpdata", output=filename, exclude=exclude)


def load_database_from_json(filename="database_dump.json"):

    # OPTIMIZE: this function is very slow. Consider speed-up options such as
    # making this function a transaction or manually writing a bulk_create. It
    # actually looks like django ORM takes up most of the time tough, and the actual
    # database queries are not the bottleneck...

    # execute the following commands to build the database
    # BUG: contenttypes gives issues because a migrated database already has these
    # set. Simply ignore this table and everything works. The contenttypes is
    # simply a table that lists all of our different models.
    call_command("loaddata", filename, exclude=["contenttypes"])


# BUG: This function isn't working as intended
# def graph_database(filename="database_graph.png"):

#     # using django-extensions, we want to make an image of all the available
#     # tables in our database as well as their relationships.

#     # This is the equivalent of running the following command:
#     #   django-admin graph_models -a -o image_of_models.png --settings=...
#     call_command("graph_models", output=filename, all_applications=True, layout="fdp")



# -*- coding: utf-8 -*-

import os

from pymatgen.io.vasp.outputs import Elfcar, Chgcar

from simmate.toolkit import Structure

from .bader import BaderAnalysis


class BaderELFAnalysis(BaderAnalysis):
    """
    Runs Bader analysis where the ELFCAR is used as the partitioning reference
    instead of CHGCAR.
    """

    command = "bader CHGCAR_empty -ref ELFCAR_empty > bader.out"
    requires_structure = True  # !!! This may change in the future
    required_files = ["CHGCAR", "ELFCAR"]

    @staticmethod
    def setup(structure: Structure, directory: str):
        """
        Bader analysis requires that a static-energy calculation be ran beforehand
        - typically using VASP. This setup involves ensuring that
        the proper files are present. In addition, it creates new CHGCAR and
        ELFCAR files with an update structure -- typically containing empty atoms.

        #### Parameters

        - `structure`:
            The structure to use when rewriting the ELFCAR and CHGCAR files.
        """

        # establish file names
        elfcar_filename = os.path.join(directory, "ELFCAR")
        chgcar_filename = os.path.join(directory, "CHGCAR")

        # Load ELFCAR + CHGCAR and replace the structure with the one we created that includes
        # empty atoms.
        elfcar = Elfcar.from_file(elfcar_filename)
        elfcar.structure = structure  # replaces the original structure in file
        elfcar.write_file(f"{elfcar_filename}_empty")
        chgcar = Chgcar.from_file(chgcar_filename)
        chgcar.structure = structure  # replaces the original structure in file
        chgcar.write_file(f"{chgcar_filename}_empty")



# -*- coding: utf-8 -*-

import os

from simmate.conftest import copy_test_files
from simmate.calculators.vasp.inputs import Incar
from simmate.calculators.vasp.error_handlers import Unconverged


def test_unconverged_electronic(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="unconverged_electronic",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")
    vasprun_filename = os.path.join(tmpdir, "vasprun.xml")

    # Confirm an error IS found when we have an unconverging xml
    error_handler = Unconverged()
    assert error_handler.check(tmpdir) == True

    # Make first attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ALGO from VeryFast to Fast"
    assert Incar.from_file(incar_filename)["ALGO"] == "Fast"

    # Make second attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ALGO from Fast to Normal"
    assert Incar.from_file(incar_filename)["ALGO"] == "Normal"

    # Make third attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "switched ALGO from Normal to All"
    assert Incar.from_file(incar_filename)["ALGO"] == "All"

    # Make fourth attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == (
        "turned on mixing with the following settings: {'ISTART': 1, 'ALGO': "
        "'Normal', 'NELMDL': -6, 'BMIX': 0.001, 'AMIX_MAG': 0.8, 'BMIX_MAG': 0.001}"
    )
    incar = Incar.from_file(incar_filename)
    assert incar["ISTART"] == 1
    assert incar["ALGO"] == "Normal"
    assert incar["NELMDL"] == -6
    assert incar["BMIX"] == 0.001
    assert incar["AMIX_MAG"] == 0.8
    assert incar["BMIX_MAG"] == 0.001

    # make sure no error is raised when the xml doesn't exist
    os.remove(vasprun_filename)
    assert error_handler.check(tmpdir) == False


def test_unconverged_ionic(tmpdir):
    copy_test_files(
        tmpdir,
        test_directory=__file__,
        test_folder="unconverged_ionic",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = os.path.join(tmpdir, "INCAR")
    # Confirm an error IS found when we have an unconverging xml
    error_handler = Unconverged()
    assert error_handler.check(tmpdir) == True

    # Make first attempt at fixing the error
    fix = error_handler.correct(tmpdir)
    assert fix == "copied the CONTCAR into the POSCAR and switched IBRION to 1"
    assert Incar.from_file(incar_filename)["IBRION"] == 1
